/** Eve Manager user interface map's component class
 * 
 * @class MapComponent
 *
 *
 * @author <a href="mailto:info@ilevia.com">info@ilevia.com</a>
 *
 */
"use strict";

function MapComponent(){
	
	/** Curent instance */
	var self = this;
	
	/** A node of component reference
	 * @public
	 * @type {Node|undefined}
	 * @default undefined
	 */
	var m_oReference = undefined;
	
	/** A Map reference
     * @public
     * @type {Map}
     * @default undefined
     */
    var m_oMapReference = undefined;
	
	/** Map html id
     * @public 
     * @type string
     * @default undefined
     */
    this.id = undefined;
	
	/** Map component's name
     * @public
     * @type string
     * @default undefined
     */
    this.name = undefined;
	
	/** Map component's scale factor
     * @public
     * @type Number
     * @default 1
     */
    this.scaleFactor = 1;
    this.scaleFactorWidth = 1;
    this.scaleFactorHeight = 1;
    this.firstTimeCreation=true;

    this.icons = new Array();

    this.iconChanged=false; // to true when the user change the icon on the map

    this.uiReadOnly=false; // EVE remote disable action on this element
    this.hideIconComponent = "never";
    this.hideIconComponentValue = "1";
    this.disableIconComponent="never";
    this.disableIconComponentValue="1";

    this.uiDisableCircle=false; // it can be useful to remove the circle on EVE Remote


    /** View mode
     * @private
     * @type Integer
     * @default undefined
     */
    this.viewmode = 0;
    this.viewmode2 = 0;
    this.viewmode3 = 0;
    this.viewmode4 = 0;
    this.viewmode5 = 0;

    /** Text alignment
     * @private
     * @type Integer
     * @default undefined
     */
    this.textalign = 0;
    
    /** Map component's position
     * @public
     * @type Object
     * @default { x:0, y:0 }
     */
    this.position = { x:0, y:0 };

    this.textFormat="";


    /** Set reference node component
     *
     */
    this.setReference = function(oNode){
        m_oReference = oNode;
        return m_oReference;
    };
    
    /** Get reference node component
     *
     */
    this.getReference = function(){
        return m_oReference;
    };
    
    /** Set reference map
     *
     */
    this.setReferenceMap = function(oMap){
        if(!oMap instanceof Map){
            console.error("Not map instance");
            return false;
        }
        m_oMapReference = oMap;
        return m_oMapReference;
    };
    
    /** Get reference map
     *
     */
    this.getReferenceMap = function(){
        return m_oMapReference;
    };
	
	/** JSON serialization
     *
     */
    this.encodeJSON = function(){
        //Set instance variables
        if (self.position.x==null) self.position.x=0;

        if (self.position.y==null) self.position.y=0;

        var aIcons = self.icons;


        if (!self.iconChanged)
        {
            var oComponent = self.getReference().getData();
            aIcons = oComponent.getParamVal("icons"); // prendo le icone dal componente originale

        }

        var oSerialized = {
            name: self.name,
            scaleFactor: self.scaleFactor,
            scaleFactorWidth: self.scaleFactorWidth,
            scaleFactorHeight: self.scaleFactorHeight,
            viewmode: self.viewmode,
            viewmode2: self.viewmode2,
            viewmode3: self.viewmode3,
            viewmode4: self.viewmode4,
            viewmode5: self.viewmode5,
            icons: aIcons,
            iconChanged: self.iconChanged,
            position: self.position,
            reference: m_oReference.getData().UId,
            uiReadOnly: self.uiReadOnly,
            hideIconComponent:self.hideIconComponent,
            hideIconComponentValue:self.hideIconComponentValue,
            disableIconComponent:self.disableIconComponent,
            disableIconComponentValue:self.disableIconComponentValue,
            uiDisableCircle : self.uiDisableCircle,
            firstTimeCreation: self.firstTimeCreation,
            textFormat:self.textFormat

        };
        return oSerialized;
    };
	
	/** Overwrite toString() method
     * @returns {string} String that describe UIComponent object
     */
    this.toString = function(){
    	return "MapComponent - Component name: "+ self.name +"; reference: "+ m_oReference.toString() + ";";
    };
}