﻿/** Model: Business logic singleton class
 *
 * @class Model
 *
 *
 * @author <a href="mailto:info@ilevia.com">info@ilevia.com</a>
 *
 * 
 * @requires jQuery framework
 */
//"use strict";

function Model(){
	//Build singleton instance
	if ( arguments.callee._singletonInstance ){
        console.warn("Singleton instance used");
        return arguments.callee._singletonInstance;
    }
        
    arguments.callee._singletonInstance = this;


    
	/** Instance variables (private) */
	var self = this;
	
	/** Initialized method
     * 
     */
    var init = function(){
        //--console.log("%cModel class initialization","color:green");
        
        //Create gateways root node
        m_oGateways = new Node({ id: "mygateways", m_Data:{  name: "root" }});
        
        //Create components root node
        m_oComponents = new Node({ id: "componentscontainer", m_Data:{  name: "root" }});
        
        //Create ui-components root node
        m_oUIComponents = new Node({ id: "myinterface", m_Data:{  name: "root" }});

        //Create users root node
        m_oUsers = new Node();

        //Create users root node
        m_oRelations = new Node();

        m_oStatus=new AppStatus();

        //Load and fill protocols combo box from "/protocols/declare.json"
        /*$.ajax({ dataType: "json", url: "./declarations/protocols.json", async: false }).success(function(aData){
            self.protocols = aData;
            for(var i=0; i<aData.length; ++i){ self.protocolsLabel[aData[i].id] = aData[i].label; }
        });*/

    };
	
	
	/** Loaded User
     * @private
     * @type User
     * @default undefined
     */
    //var m_refLoadUser = undefined;
	
	/** Interface building selected element
	 * @private
	 * @type Node
	 * @default undefined
	 */
	var m_refSelUIComponent = undefined;
	
	/** Created components selected component
	 * @private
	 * @type Node
	 * @default undefined
	 */
	var m_refSelComponent = undefined;
	
	/** Selected gateway (at least one gateway need to be always selected, ghost by default)
	 * @private
	 * @type Node
	 * @default undefined
	 */
	var m_refSelGateway;

    /** Selected user
     * @private
     * @type Node
     * @default undefined
     */
    var m_refSelUser;

    /** Users root node reference (tree structure)
     * @private
     * @type Node
     * @default new Node({ id: "level-1", m_Data:{  name: "root" }});
     */
    var m_oUsers;


    var m_oStatus;

	/** Interface tab root node reference (tree structure)
	 * @private
	 * @type Node
	 * @default new Node({ id: "level-1", m_Data:{  name: "root" }});
	 */
	var m_oUIComponents;
	
	/** My components tab root node reference (tree structure)
	 * @private
	 * @type Node
	 * @default new Node({ id: "componentscontainer", m_Data:{  name: "root" }});
	 */
	var m_oComponents;
	
	/** My gateways tab root node reference (tree structure)
	 * @private
	 * @type Node
	 * @default new Node({ id: "mygateways", m_Data:{  name: "root" }});
	 */
	var m_oGateways;

    /** Relations root node reference (tree structure)
     * @private
     * @type Node
     * @default new Node();
     */
    var m_oRelations;
	
	
	/** Application's version
     * @public
     * @type String
     */
    this.version = undefined;
    
    /** Application's name
     * @public
     * @type String
     */
    this.name = undefined;
	
	/** Eve Manager avaiable protocols (loaded in eve-manager-init.js)
	 * @private
	 * @type Array
	 * @default new Array()
	 */
	this.protocols = new Array();

    /** Eve Manager skin informations (loaded in eve-manager-init.js)
     * @private
     * @type Array
     * @default new Array()
     */
    this.skin = new Array();


	
	/** Eve Manager label of avaiable protocols (loaded in eve-manager-init.js)
	 * @public
	 * @type Array
	 * @default new Array()
	 */
	this.protocolsLabel = new Array();
	
	/** Application HID (Human Interface Device) manager
	 * @public
	 * @type HidManager
	 */
	//this.HidManager = new HidManager();
	
	
	/** Model resources manager
     * @public
     * @type ResourcesManager
     */
    this.ResourcesManager = new ResourcesManager();

    /** User html element id generator
     * @public
     * @type EveIdAssign
     */
    this.UserIdAssign = new EveIdAssign();
	
	/** Interface html element id generator
	 * @public
	 * @type EveIdAssign
	 */
	this.InterfaceIdAssign = new EveIdAssign();
	
	/** Interface User Interface Unique Id generator
	 * @public
	 * @type Indexer
	 */
	this.InterfaceUIIdAssign = new Indexer();
	
	/** My component html element id generator
	 * @public
	 * @type EveIdAssign
	 */
	this.ComponentIdAssign = new EveIdAssign();
	
	/** Components Unique Id generator
	 * @public
	 * @type Indexer
	 */
	this.ComponentUniqueIdAssign = new Indexer();
	
	/** My gateways html element id generator
	 * @public
	 * @type EveIdAssign
	 */
	this.GatewaysIdAssign = new EveIdAssign();
	
	/** Gateways Unique Id generator
	 * @public
	 * @type Indexer
	 */
	this.GatewaysUniqueIdAssign = new Indexer();

    /** Relation id generator
     * @public
     * @type EveIdAssign
     */
    this.RelationsIdAssign = new EveIdAssign();

    /** Optimization dictionary
     * @public
     * @type ComponentID
     */
    this.optimizedDictionary= new Array();



    /** Gateways methods (gateway's name is the unique identifier) */
	
	/** Get gateways root node
	 * @returns {Node|undefined} Root node of selected gateway
	 */
	this.getGatewaysRoot = function(){
		return m_oGateways;
	};
	
	/** Add gateway to application
	 * @param {object|null} oParam An object with gateway's parameters as "name" or
	 * @returns {Node} Node linked to new gateway
	 */
	this.addGateway = function(strUID){
		var strId = "gateway_"+ self.GatewaysIdAssign.getId();
		var UniqueId;
		
		if(strUID != undefined){
			//Use exited
			//console.debug("UId exixt!");
			self.GatewaysUniqueIdAssign.setHighestId(strUID);                                                //Set
			//self.GatewaysUniqueIdAssign.setHighestId(oParam.UId);												//Set
			//UniqueId = oParam.UId;
			UniqueId = strUID;
		}
		else{
			//Get new
			UniqueId = self.GatewaysUniqueIdAssign.getNextFreeId();
		}
		
		var oGateway = new Gateway();
		oGateway.UId = UniqueId;																		//Set gateway unique Id
		oGateway.id = strId;																			//Set gateway html Id
		
		var oNode = new Node(oGateway);																	//Create node component
        m_oGateways.addChild(oNode);                                                                    //Add node to parent
		return oNode;
	};
	
	/** Remove gateway
	 * @param {Node} oParam An object with gateway's parameters as "name" or
	 */
	this.removeGateway = function(oNode){
		var oGateway = oNode.getData();
		var strId = oGateway.id;
		
		//Remove child node
		if(m_oGateways.removeChild(oNode) != 0){
		    throw "Gateway node can't be removed";
			//console.error("Gateway node can't be removed");
			return;
		}
	};
	
	/** Get selected gateway
     * @returns {Node} Node of selected gateway
     */
    this.getSelectGateway = function(){
        return m_refSelGateway;
    };
	
	/** Set selected gateway
	 * @returns {Node|undefined} Node of selected gateway, undefined if any node is selected
	 */
	this.setSelectGateway = function(oNode){
		//Deselect gateway if oNode is undefined
		if(oNode == undefined){
		    self.deselectGateway();
		    return undefined;
		}
		
		//Check node instance
		if(! (oNode instanceof Node) ){
		    throw "oNode is not a Node instance";
		    //console.error("oNode is not a Node instance");
		    return undefined;
		}
		
		//Get component data
		var oGateway = oNode.getData();
		
		//Check component id existance
		if(oGateway.id == undefined){
		    throw "Selected gateway has not any Id";
		    //console.error("Selected gateway has not any Id");
		    return undefined;
		}
				
		//Deselect previous gateway
		self.deselectGateway();
		
		m_refSelGateway = oNode;																	//Set selected component reference to selected node
		
		return oNode;
	};
	
	/** Remove selection from selected gateway
	 * @returns {undefined} Anything
	 */
	this.deselectGateway = function(){
	    //var oNode = m_refSelGateway;
		m_refSelGateway = undefined;                                                                //Set selected gateway to null
	};
	
	/** Get gateways root node
     * @returns {Node|undefined} Root node of selected gateway
     */
    this.getGatewayByUid = function(strUId,oRoot){
        if(strUId == undefined){
            return undefined;
        }
        oRoot = (oRoot != undefined) ? oRoot : m_oGateways;                                           //Set root node of sub-tree 
        var rootUId = oRoot.getData().UId;                                                              //Get unique Id (UId)
        
        if(rootUId == strUId){
            return oRoot;
        }
        else{
            var oNode = undefined;
            for(var i=0; i<oRoot.children.length; ++i){
                oNode = self.getGatewayByUid(strUId,oRoot.children[i]);
                if(oNode != undefined){ return oNode; }
            }
            return oNode;
        }
    };
	
	
    /** Components methods */
    
    /** Get components root node
     * @returns {Node} Root node of components
     */
    this.getComponentsRoot = function(){
        return m_oComponents;
    };
    
    /** Create a new Node with component as data
     * @param strUID Component UID if exist
     * @param oParent The parent node of new node. If == undefined than the new node will be added to the root
     * @returns If successed new oNode otherwise undefined
     */
    this.addComponent = function(strUID,oParent){
        var strId = "component_"+ self.ComponentIdAssign.getId();
        var UniqueId;

        //Assign Unique Id
        if(strUID != undefined){
            //Use exited
            //console.debug("UId exixt!");
            self.ComponentUniqueIdAssign.setHighestId(strUID);
            //UniqueId = oParam.UId;
            UniqueId = strUID;
        }
        else{
            //Get new
            UniqueId = self.ComponentUniqueIdAssign.getNextFreeId();
        }

        //Component creation (component initialization is inside model function)
        var oComponent = new Component();
        oComponent.UId = UniqueId;                                                                      //Set component unique Id
        oComponent.id = strId;                                                                          //Set component html Id

        var oNode = new Node(oComponent);                                                               //Create node component

        if(oParent == undefined){
            //oParent = m_oComponents;
            oParent = self.getComponentsRoot();
        }

        self.optimizedDictionary[oComponent.UId]=oNode;

        oParent.addChild(oNode);
        
        return oNode;
    };
    
    /** Remove Node
     * @param oNode The node component that will be removed.
     * @returns If successed 0 otherwise 1
     */
    this.removeComponent = function(oNode){

        if(oNode.isRoot()){
            throw "Can not remove root node";
            //console.error("Can not remove root node");
            return false;
        }
        //console.log(oNode);
        var oParent = oNode.getParent();
        var oComponent = oNode.getData();


        if(oParent == undefined){
            throw "Can not remove node from undefined parent";
            //console.error("Can not remove node");
            return false;
        }
        
        if(oParent.removeChild(oNode) != 0){
            throw "Can not remove node";
            //console.error("Can not remove node");
            return false;
        }
        delete self.optimizedDictionary[oComponent.UId];
        return true;
    };
    
    /** Set selected node
     * @param oNode node object
     */
    this.setSelectComponent = function(oNode){
        //Check node instance
        if(! (oNode instanceof Node)){
            console.error("oNode is not a Node instance");
            return undefined;
        }
        
        //Get component data
        var oComponent = oNode.getData();
        
        //Check node instance
        if(! (oComponent instanceof Component) ){
            console.error("oComponent is not a Component instance");
            return undefined;
        }
        
        //self.deselectComponent();                                                                     //Deselect all components
        
        m_refSelComponent = oNode;                                                                      //Set selected component reference to selected node
        
        return oNode;
    };
    
    /** Return selected component in Eve Manager instance
     * @returns {Node|undefined} Selected component node
     */
    this.getSelectComponent = function(){
        return m_refSelComponent;
    };
    
    /** Deselect all components
     * @returns {Node|undefined} Node going to be deselected
     */
    this.deselectComponent = function(){
        var oNode = m_refSelComponent;
        m_refSelComponent = undefined;                                                                  //Set selected component reference to undefined
        return oNode;
    };
    
    /** Get component by unique Id (UId) inside a sub-tree structure with input node as root
     * @param {string} strId Unique id for match
     * @param {Node|undefined} oRoot Root node of sub-tree to find Node by id
     * @returns {Node|undefined} Node corresponding to id
     */
    this.getComponentByUid = function(strUId,oRoot){
    	//console.log("Model - GetComponentsByUID");

        if (self.optimizedDictionary[strUId]!=undefined)
        {
            return self.optimizedDictionary[strUId];
        }

        if(strUId == undefined){
            return undefined;
        }
        oRoot = (oRoot != undefined) ? oRoot : m_oComponents;                                           //Set root node of sub-tree 
        var rootUId = oRoot.getData().UId;                                                              //Get unique Id (UId)
        
        //console.debug("%c"+rootUId,"color:purple;");
        //console.debug("%cstrUId: %c"+ strUId,"font-size:16px;","font-size:16px; color:orange;");
        //console.debug("%crootUId: %c"+ rootUId,"font-size:16px;","font-size:16px; color:orange;");
        if(rootUId == strUId){
            //console.debug(oRoot.getData().getParamVal("name"));
            return oRoot;
        }
        else{
            //console.debug(oRoot.getData().getParamVal("name"));
            //console.debug(oRoot.getData());
            var oNode = undefined;
            for(var i=0; i<oRoot.children.length; ++i){
                oNode = self.getComponentByUid(strUId,oRoot.children[i]);
                if(oNode != undefined)
                {

                    return oNode;
                }
            }
            
            //return undefined;
            return oNode;
        }
    };
    
    
    /** Users methods */

    /** Get users root node
     * @returns {Node|undefined} Root node
     */
    this.getUsersRoot = function(){
        return m_oUsers;
    };

    /** Add user
     * @returns {Node} Root node of components
     */
    this.addUser = function(strName){

        //Check
        if(strName == undefined || typeof strName != "string"){
            throw "Invalid strName parameter";
        }

        var strId = "user_"+ self.UserIdAssign.getId();             //Set id for html
        //Check name existance
        var bFind = false;
        for(var i=0,l=m_oUsers.children.length; i<l; ++i){
            if(m_oUsers.children[i].getData().name == strName){
                bFind = true;
                break;
            }
        }

        if(bFind){
            strName=strId;
            //throw "User name already exist";
        }

        // User Interface Component creation (component initialization is inside model function)
        var oUser = new User();

        //Set user html Id
        oUser.id = strId;
        oUser.name = strName;

        var oNode = new Node(oUser);                                        //Create node component
        m_oUsers.addChild(oNode);

        return oNode;
    };

    /** Remove user
     *
     */
    this.removeUser = function(oNode){

        var oUser = oNode.getData();
        var strId = oUser.id;

        //Remove child node
        if(m_oUsers.removeChild(oNode) != 0){
            throw "User node can't be removed";
            //console.error("Gateway node can't be removed");
            return;
        }
    };

    /** Set selected node. Re-initialize interface root-node
     * @param oNode node object
     */
    this.setSelectUser = function(oNode){
        //Check node instance
        if(! (oNode instanceof Node) ){
            throw "oNode is not a Node instance";
            console.error("oNode is not a Node instance");
            return undefined;
        }

        /*function emptyTree(oNode){
            if(oNode.children.length > 0){
                for(var i=0; i<oNode.children.length; ++i){
                    emptyTree( oNode.children[i] );
                }
            }

            for(var i=0; i<oNode.children.length; ++i){
                oNode.removeChild( oNode.children[i] );
            }

        };*/

        // Reset interface root node
        //emptyTree(m_oUIComponents);

        // Reset interface selected node
        m_refSelUIComponent = undefined;

        //Check node instance
        if(! (oNode.getData() instanceof User) ){
            throw "oNode does not contain a UIComponent instance";
            console.error("oNode does not contain a UIComponent instance");
            return undefined;
        }

        //Set selected UI node
        m_refSelUser = oNode;

        return m_refSelUser;
    };

    /** Get selected user
     * @returns {Node} Selected user interface component
     */
    this.getSelectUser = function(){
        return m_refSelUser;
    };


    
    //this.removeUser = function(oUser){}
    
    //this.getUserByName = function(strName){}
    
    
	/** Interface methods */
	
	/** Get components root node
	 * @returns {Node} Root node of components
	 */
	this.getUIComponentsRoot = function(){
		return m_oUIComponents;
	};
	
	/** Add element to interface
	 * @param {Node} oNode Node to add
  	 * @param {Node|undefined} oParent Parent node to add children node
	 * @param {jQuery} $element jQuery element
	 */
	this.addUIComponent = function(strUID,oParent,UILevel){
	    var strId = "interface_"+ self.InterfaceIdAssign.getId();                  //Set id for html linked component
        var UniqueId;
	    console.log("add ui model");
	    //Assign Unique Id
        if(strUID != undefined){
            //Use exited
            //console.debug("UId exixt!");
            self.InterfaceUIIdAssign.setHighestId(strUID);
            //UniqueId = oParam.UId;
            UniqueId = strUID;
        }
        else{
            //Get new
            UniqueId = self.InterfaceUIIdAssign.getNextFreeId();
        }
	    
	    // User Interface Component creation (component initialization is inside model function)
        var oUIComponent = new UIComponent();
        
        //oUIComponent.setUIdGenerator(self.InterfaceUIIdAssign);                     //New
        
        oUIComponent.UId = UniqueId;           	                                 //Set component unique Id
		oUIComponent.UILevel=UILevel;
        oUIComponent.id = strId;                                                   //Set component html Id
        
        var oNode = new Node(oUIComponent);                                        //Create node component
        
        if(oParent == undefined){
            oParent = self.getUIComponentsRoot();
            //oParent = m_oUIComponents;
        }
        oParent.addChild(oNode);
        
        return oNode;
	};
	
	/** Remove user interface component
	 * @param {Node} oParam An object with gateway's parameters as "name" or
	 */
	this.removeUIComponent = function(oNode){
	    if(oNode.isRoot()){
            console.error("Can not remove root node");
            return 1;
        }
        
        var oParent = oNode.getParent();
        
        if(oParent.removeChild(oNode) != 0){
            console.error("Can not remove node");
            return 2;
        }
        
        return 0;
	};
	
	/** Set selected node
	 * @param oNode node object
	 */
	this.setSelectUIComponent = function(oNode){
		//Check node instance
		if(! (oNode instanceof Node) ){
		    console.error("oNode is not a Node instance");
		    return undefined;
		}
		
		//Check node instance
        if(! (oNode.getData() instanceof UIComponent) ){
            console.error("oNode does not contain a UIComponent instance");
            return undefined;
        }
		
		//Set selected UI node
		m_refSelUIComponent = oNode;
		
		return m_refSelUIComponent;
	};
	
	/** Get selected node
	 * @returns {Node} Selected user interface component  
	 */
	this.getSelectUIComponent = function(){
		return m_refSelUIComponent;
	};
	
	/** Deselect interface element
	 * @return {Node|undefined} Selected node if is defined
	 */
	this.deselectUIComponent = function(){
		if(m_refSelUIComponent == undefined){
		    return undefined;
		}
		var oNode = m_refSelUIComponent;
		
		m_refSelUIComponent = undefined;
		return oNode;
	};

	/** Get UI component by unique Id (UId) inside a sub-tree structure with input node as root
     * @param {string} strId Unique id for match
     * @param {Node|undefined} oRoot Root node of sub-tree to find Node by id
     * @returns {Node|undefined} Node corresponding to id
     */
    this.getUIComponentByUid = function(strUId,oRoot){
        
        if(strUId == undefined){
            return undefined;
        }
        oRoot = (oRoot != undefined) ? oRoot : m_oUIComponents;                                         //Set root node of sub-tree 
        var rootUId = oRoot.getData().UId;                                                              //Get unique Id (UId)
        
        //console.debug("%c"+rootUId,"color:purple;");
        //console.debug("%cstrUId: %c"+ strUId,"font-size:16px;","font-size:16px; color:orange;");
        //console.debug("%crootUId: %c"+ rootUId,"font-size:16px;","font-size:16px; color:orange;");
        if(rootUId == strUId){
            //console.debug(oRoot.getData().getParamVal("name"));
            return oRoot;
        }
        else{
            //console.debug(oRoot.getData().getParamVal("name"));
            //console.debug(oRoot.getData());
            var oNode = undefined;
            for(var i=0; i<oRoot.children.length; ++i){
                oNode = self.getUIComponentByUid(strUId,oRoot.children[i]);
                if(oNode != undefined){
                    return oNode;
                }
            }
            
            return oNode;
        }
    };
    
    /** Get UI component by unique Id (UId) inside a sub-tree structure with input node as root
     * @param {string} strId Unique id for match
     * @param {Node|undefined} oRoot Root node of sub-tree to find Node by name
     * @returns {Node|undefined} Node corresponding to id
     */
    this.getUIComponentByName = function(strName,oRoot){
        
        if(strName == undefined){
            return undefined;
        }
        oRoot = (oRoot != undefined) ? oRoot : m_oComponents;    //Set root node of sub-tree 
        var rootUId = oRoot.getData();      //Get unique Id (UId)
        
        if(strName == oRoot.getData().m_Data.name){
            return oRoot;
        }
        else{
            var oNode = undefined;
            for(var i=0; i<oRoot.children.length; ++i){
                oNode = self.getUIComponentByName(strName,oRoot.children[i]);
                if(oNode != undefined){ return oNode; }
            }
            
            //return undefined;
            return oNode;
        }
    };
    
    /** Get UI component by unique userData inside a sub-tree structure with input node as root
     * @param {string} strId Unique id for match
     * @param {Node|undefined} oRoot Root node of sub-tree to find Node by name
     * @returns {Node|undefined} Node corresponding to id
     */
    this.getUIComponentByUserData = function(strName,oRoot){
        
        if(strName == undefined){
            return undefined;
        }
        oRoot = (oRoot != undefined) ? oRoot : m_oComponents;    //Set root node of sub-tree 
        var rootUId = oRoot.getData();      //Get unique Id (UId)
        
        if(strName == oRoot.getData().m_Data["ghost/userdata"]){
            return oRoot;
        }
        else{
            var oNode = undefined;
            for(var i=0; i<oRoot.children.length; ++i){
                oNode = self.getUIComponentByUserData(strName,oRoot.children[i]);
                if(oNode != undefined){ return oNode; }
            }
            
            //return undefined;
            return oNode;
        }
    };

    /** Relations */

    /** Get relations root node
     * @returns {Node} Root node of components
     */
    this.getRelationsRoot = function(){
        return m_oRelations;
    };

    this.getStatus = function(){
        return m_oStatus;
    };

    /** Add relations
     *
     */
    this.addRelation = function(){
        var strId = "relation_"+ self.RelationsIdAssign.getId();                  //Set id for html linked component

        // User Interface Component creation (component initialization is inside model function)
        var oRelation = new Relation();
        oRelation.id = strId;

        var oNode = new Node(oRelation);                                        //Create node component

        m_oRelations.addChild(oNode);

        return oNode;
    };

    /** Add relations */
    this.removeRelation = function(oNode){
        m_oRelations.removeChild(oNode);
        return true;
    };

    /** Reset memory Interface
     *
     */
    /*this.resetInterface = function(){
        // Reset root node
        m_oUIComponents = undefined;

        // Reset selected node
        m_refSelUIComponent = undefined;
    };*/

	/** Reset memory area
	 * 
	 */
	this.resetModel = function(){
	    // Reset root nodes
	    m_oGateways = undefined;
	    m_oComponents = undefined;
	    m_oUIComponents = undefined;
        m_oUsers = undefined;
	    
	    // Reset selected nodes
	    m_refSelGateway = undefined;
	    m_refSelComponent = undefined;
	    m_refSelUIComponent = undefined;
        m_refSelUser = undefined;

        //Reset UId assigner
        self.InterfaceUIIdAssign = new Indexer();
        self.ComponentUniqueIdAssign = new Indexer();
        self.GatewaysUniqueIdAssign = new Indexer();

	    //Initialize Model
	    init();
	};
	
	
	/** Initialize Model */
	init();
}