/**
 * Created by blr on 07/05/15.
 */
"use strict";

var VERSION = "v.1.6.1";
var NAME = "Import VIMAR By Me";
var $ = jQuery;

try {
    console.log("%cStart " + NAME, "color:DarkSlateBlue");
    create_gui();
    console.log("%cEnd " + NAME, "color:DarkSlateBlue");
} catch (err) {
    console.error(NAME + " error\n" + err);
    alert(NAME + " error\n" + err);
}

function vimar_main
(
    gtwAddress,
    fileByme,
    fileKnx,
    createFeedback,
    startFeedbackAddress
)
{
    try {
        if(fileByme == ""){
            throw "Invalid Vimar By-Me xml file";
        }

        var xmlDoc = $.parseXML(QT.readAbsoluteFile(fileByme));
        if (xmlDoc == undefined) {
            throw "Empty or invalid Vimar By-Me xml file";
        }

        var csv = { file:"",data:"" };
        if (fileKnx != "") {
            csv.file = fileKnx;
            csv.data = build_csv_row("NAME","DPT","WRITE","READ","VALUE");
        }

        var feedbacks = { create: createFeedback, address: 0, relations: {} }
        if (feedbacks.create) {
            // Check address sanity
            feedbacks.address = groupAddressStringToValue(startFeedbackAddress);
            if (feedbacks.address == -1) {
                throw "Invalid start group address [" + startFeedbackAddress + "]";
            }
        }
        
        // Check main root node
        var xmlRoot = xmlDoc.childNodes[0];
        if (xmlRoot.nodeName != "plant") {
            throw "Invalid Xml File, main node plant not found.";
        }

        // Get configurator offset
        var xmlConfigurator = xmlRoot.getElementsByTagName("configurator")[0];
        if (xmlConfigurator == undefined) {
            throw "configurator node not found";
        }

        var visitor = new blrComponentsVisitor();
        var configuratorOffset = parseInt(xmlConfigurator.getAttribute("offset").toString(16), 16);
        var xmlApplications = xmlConfigurator.getElementsByTagName("application");

        var gateways = getGateways(gtwAddress);
        var scenaries = new Array();
        var scenariesName = new Array();
        var scenariesType = new Array(); // true if local, false if global

        for (var i = 0; i < xmlApplications.length; ++i) {
            var xmlApplication = xmlApplications[i];
            var id = parseInt(xmlApplication.getAttribute("id"), 10);
            switch(id){
                case 0: { // Automation
                    var xmlIndices = xmlApplication.getElementsByTagName("index");
                    if (xmlIndices.length > 0) {
                        buildAutomation(visitor, xmlIndices, gateways, scenaries, feedbacks, csv);
                    }
                } break;
                case 1: { // Clima
                    var xmlIndices = xmlApplication.getElementsByTagName("index");
                    if (xmlIndices.length > 0) {
                        buildClima(visitor, xmlIndices, gateways, scenaries, feedbacks, csv);
                    }
                } break;
                case 2: { // Carichi
                    var xmlIndices = xmlApplication.getElementsByTagName("index");
                    if (xmlIndices.length > 0) {
                        buildPower(visitor, xmlIndices, gateways, scenaries, feedbacks, csv);
                    }
                } break;
                case 4: { // Global scenaries
                    var xmlIndices = xmlApplication.getElementsByTagName("index");
                    if (xmlIndices.length > 0) {
                        buildScenariesMaps(xmlIndices, scenaries, scenariesName, scenariesType, false, feedbacks, csv);
                    }
                } break;
                case 5: { // Local scenaries
                    var xmlIndices = xmlApplication.getElementsByTagName("index");
                    if (xmlIndices.length > 0) {
                        buildScenariesMaps(xmlIndices, scenaries, scenariesName, scenariesType, true, feedbacks, csv);
                    }
                } break;
                case 9: { // Media
                    var xmlIndices = xmlApplication.getElementsByTagName("index");
                    if (xmlIndices.length > 0) {
                        buildMedia(visitor, xmlIndices, gateways, scenaries, feedbacks, csv);
                    }
                } break;
            }
        }

        // Build scenaries
        blr.executionStack.push(
            function (in_args, out_args) {
                out_args = buildScenaries(
                    in_args.visitor,
                    in_args.gateways,
                    in_args.scenaries,
                    in_args.scenariesName,
                    in_args.scenariesType,
                    in_args.configuratorOffset,
                    in_args.feedbacks,
                    in_args.csv
                );
            },
            {
                visitor:visitor,
                gateways: gateways,
                scenaries: scenaries,
                scenariesName: scenariesName,
                scenariesType: scenariesType,
                configuratorOffset: configuratorOffset,
                feedbacks: feedbacks,
                csv: csv
            }
        );

        if(fileKnx != ""){
            blr.executionStack.push(
                function (in_args, out_args) {
                    QT.writeAbsoluteFile(in_args.csv.file, in_args.csv.data);
                },
                {
                    csv: csv
                }
            );
        }

        blr.executionStack.execute(10, true);
    } catch (err) {
        blr.executionStack.cancel();
        console.error(NAME + " error\n" + err);
        alert(NAME + " error\n" + err);
    }
}

function buildMedia
(
    visitor,
    xmlIndices,
    gateways,
    scenaries,
    feedbacks,
    csv
) {
    var groups = {
        media: undefined
    };

    for (var i = 0; i < xmlIndices.length; ++i) {
        var xmlIndex = xmlIndices[i];
        var xmlDatapoints = xmlIndex.getElementsByTagName("datapointtypes")[0].childNodes;
        if (xmlDatapoints == undefined) {
            console.log("datapointtypes node not found for index:" + xmlIndex.getAttribute("index"));
            continue;
        }

        if (xmlDatapoints.length == 0) {
            console.log("datapointtypes length == 0 for index:" + xmlIndex.getAttribute("index"));
            continue;
        }

        var xmlDevices = xmlIndex.getElementsByTagName("device");
        if (xmlDevices == undefined) {
            console.log("device node not found for index:" + xmlIndex.getAttribute("index"));
            continue;
        }

        if (xmlDevices.length == 0) {
            console.log("device length == 0 for index:" + xmlIndex.getAttribute("index"));
            continue;
        }

        var xmlDevice = undefined;
        var xmlFunctionalBlock = undefined;
        for (var d = 0; d < xmlDevices.length; ++d) {
            xmlDevice = xmlDevices[d];
            var main = undefined;
            var xmlFunctionalBlocks = xmlDevice.getElementsByTagName("functionalblock");
            for (var f = 0; f < xmlFunctionalBlocks.length; ++f) {
                xmlFunctionalBlock = xmlFunctionalBlocks[f];
                main = xmlFunctionalBlock.getAttribute("main");
                if (main == undefined) {
                    xmlFunctionalBlock = undefined;
                    continue;
                }

                if (main == "True") {
                    break;
                }
            }

            if (main == undefined) {
                xmlDevice = undefined;
                xmlFunctionalBlock = undefined;
                continue;
            }

            if (main == "True") {
                break;
            }
        }

        if (xmlDevice == undefined || xmlFunctionalBlock == undefined) {
            continue;
        }

        if (groups.media == undefined) {
            groups.media = application.addComponent({
                type: "group",
                name: "Vimar-" + _D("Media")
            });
        }

        blr.executionStack.push(
            function (in_args, out_args) {
                out_args = create_media_component(
                    in_args.visitor,
                    in_args.xmlIndex,
                    in_args.xmlDatapoints,
                    in_args.xmlDevice,
                    in_args.xmlFunctionalBlock,
                    in_args.groups,
                    in_args.gateways,
                    in_args.scenaries,
                    in_args.feedbacks,
                    in_args.csv
                );
            },
            {
                visitor: visitor,
                xmlIndex: xmlIndex,
                xmlDatapoints: xmlDatapoints,
                xmlDevice: xmlDevice,
                xmlFunctionalBlock: xmlFunctionalBlock,
                groups: groups,
                gateways: gateways,
                scenaries: scenaries,
                feedbacks: feedbacks,
                csv: csv
            }
        );
    }

    if (groups.media != undefined) {
        blr.executionStack.push(
            function (in_args, out_args) {
                var aRemove = new Array();

                var aChildren = in_args.group.children;
                for (var i = 0; i < aChildren.length; ++i) {
                    if (aChildren[i].children.length == 0) {
                        aRemove.push(aChildren[i]);
                    }
                }

                for (var i = 0; i < aRemove.length; ++i){
                    application.removeComponent(aRemove[i]);
                }

                if (in_args.group.children.length == 0) {
                    application.removeComponent(in_args.group);
                }
            },
            {
                group: groups.media
            }
        );
    }

}

function buildClima
(
    visitor,
    xmlIndices,
    gateways,
    scenaries,
    feedbacks,
    csv
)
{
    var groups = {
        clima: undefined
    };

    for (var i = 0; i < xmlIndices.length; ++i) {
        var xmlIndex = xmlIndices[i];
        var xmlDatapoints = xmlIndex.getElementsByTagName("datapointtypes")[0].childNodes;
        if (xmlDatapoints == undefined) {
            console.log("datapointtypes node not found for index:" + xmlIndex.getAttribute("index"));
            continue;
        }

        if (xmlDatapoints.length == 0) {
            console.log("datapointtypes length == 0 for index:" + xmlIndex.getAttribute("index"));
            continue;
        }

        var xmlDevices = xmlIndex.getElementsByTagName("device");
        if (xmlDevices == undefined) {
            console.log("device node not found for index:" + xmlIndex.getAttribute("index"));
            continue;
        }

        if (xmlDevices.length == 0) {
            console.log("device length == 0 for index:" + xmlIndex.getAttribute("index"));
            continue;
        }

        var xmlDevice = undefined;
        var xmlFunctionalBlock = undefined;
        for (var d = 0; d < xmlDevices.length; ++d) {
            xmlDevice = xmlDevices[d];
            var main = undefined;
            var xmlFunctionalBlocks = xmlDevice.getElementsByTagName("functionalblock");
            for (var f = 0; f < xmlFunctionalBlocks.length; ++f) {
                xmlFunctionalBlock = xmlFunctionalBlocks[f];
                main = xmlFunctionalBlock.getAttribute("main");
                if (main == undefined) {
                    xmlFunctionalBlock = undefined;
                    continue;
                }

                if (main == "True") {
                    break;
                }
            }

            if (main == undefined) {
                xmlDevice = undefined;
                xmlFunctionalBlock = undefined;
                continue;
            }

            if (main == "True") {
                break;
            }
        }

        if (xmlDevice == undefined || xmlFunctionalBlock == undefined) {
            continue;
        }

        if (groups.clima == undefined) {
            groups.clima = application.addComponent({
                type: "group",
                name: "Vimar-" + _D("Clima")
            });
        }

        blr.executionStack.push(
            function (in_args, out_args) {
                out_args = create_clima_component(
                    in_args.visitor,
                    in_args.xmlIndex,
                    in_args.xmlDatapoints,
                    in_args.xmlDevice,
                    in_args.xmlFunctionalBlock,
                    in_args.groups,
                    in_args.gateways,
                    in_args.scenaries,
                    in_args.feedbacks,
                    in_args.csv
                );
            },
            {
                visitor: visitor,
                xmlIndex: xmlIndex,
                xmlDatapoints: xmlDatapoints,
                xmlDevice: xmlDevice,
                xmlFunctionalBlock: xmlFunctionalBlock,
                groups: groups,
                gateways: gateways,
                scenaries: scenaries,
                feedbacks: feedbacks,
                csv: csv
            }
        );
    }

    if (groups.clima != undefined) {
        blr.executionStack.push(
            function (in_args, out_args) {
                if (in_args.group.children.length == 0) {
                    application.removeComponent(in_args.group);
                }
            },
            {
                group: groups.clima
            }
        );
    }
}

function buildPower
(
    visitor,
    xmlIndices,
    gateways,
    scenaries,
    feedbacks,
    csv
)
{
    var groups = {
        power: undefined
    };

    for (var i = 0; i < xmlIndices.length; ++i) {
        var xmlIndex = xmlIndices[i];
        var xmlDatapoints = xmlIndex.getElementsByTagName("datapointtypes")[0].childNodes;
        if (xmlDatapoints == undefined) {
            console.log("datapointtypes node not found for index:" + xmlIndex.getAttribute("index"));
            continue;
        }

        if (xmlDatapoints.length == 0) {
            console.log("datapointtypes length == 0 for index:" + xmlIndex.getAttribute("index"));
            continue;
        }

        var xmlDevices = xmlIndex.getElementsByTagName("device");
        if (xmlDevices == undefined) {
            console.log("device node not found for index:" + xmlIndex.getAttribute("index"));
            continue;
        }

        if (xmlDevices.length == 0) {
            console.log("device length == 0 for index:" + xmlIndex.getAttribute("index"));
            continue;
        }

        var xmlDevice = undefined;
        var xmlFunctionalBlock = undefined;
        for (var d = 0; d < xmlDevices.length; ++d) {
            xmlDevice = xmlDevices[d];
            var main = undefined;
            var xmlFunctionalBlocks = xmlDevice.getElementsByTagName("functionalblock");
            for (var f = 0; f < xmlFunctionalBlocks.length; ++f) {
                xmlFunctionalBlock = xmlFunctionalBlocks[f];
                main = xmlFunctionalBlock.getAttribute("main");
                if (main == undefined) {
                    xmlFunctionalBlock = undefined;
                    continue;
                }

                if (main == "True") {
                    break;
                }
            }

            if (main == undefined) {
                xmlDevice = undefined;
                xmlFunctionalBlock = undefined;
                continue;
            }

            if (main == "True") {
                break;
            }
        }

        if (xmlDevice == undefined || xmlFunctionalBlock == undefined) {
            continue;
        }

        if (groups.power == undefined) {
            groups.power = application.addComponent({
                type: "group",
                name: "Vimar-" + _D("Power")
            });
        }

        blr.executionStack.push(
            function (in_args, out_args) {
                out_args = create_power_component(
                    in_args.visitor,
                    in_args.xmlIndex,
                    in_args.xmlDatapoints,
                    in_args.xmlDevice,
                    in_args.xmlFunctionalBlock,
                    in_args.groups,
                    in_args.gateways,
                    in_args.scenaries,
                    in_args.feedbacks,
                    in_args.csv
                );
            },
            {
                visitor: visitor,
                xmlIndex: xmlIndex,
                xmlDatapoints: xmlDatapoints,
                xmlDevice: xmlDevice,
                xmlFunctionalBlock: xmlFunctionalBlock,
                groups: groups,
                gateways: gateways,
                scenaries: scenaries,
                feedbacks: feedbacks,
                csv: csv
            }
        );
    }

    if (groups.power != undefined) {
        blr.executionStack.push(
            function (in_args, out_args) {
                if (in_args.group.children.length == 0) {
                    application.removeComponent(in_args.group);
                }
            },
            {
                group: groups.power
            }
        );
    }

}

function buildScenaries
(
    visitor,
    gateways,
    scenaries,
    scenariesName,
    scenariesType,
    configuratorOffset,
    feedbacks,
    csv
)
{
    var groupScenaries = undefined;
    if (Object.keys(scenariesName).length > 0) {
        groupScenaries = application.addComponent({
            type: "group",
            name: "Vimar-" + _D("Scenaries")
        });
    } else {
        return;
    }

    for (var key in scenaries) {
        var aData = scenaries[key];
        scenaries[key] = aData.filter(function (elem, pos) {
            return aData.indexOf(elem) == pos;
        });

        var value = (parseInt(key.toString(16), 16)) - configuratorOffset;
        if (!scenariesType[key]) { 
            value = (parseInt(key.toString(16), 16)) - 32;
        }

        create_scenario_component
        (
            visitor,
            groupScenaries,
            gateways,
            scenariesName[key],
            key,
            value,
            scenaries[key].join(","),
            feedbacks,
            csv
        );
    }

    if (groupScenaries.children.length == 0) {
        application.removeComponent(groupScenaries);
    }
}


function buildScenariesMaps
(
    xmlIndices,
    scenaries,
    scenariesName,
    scenariesType,
    bGlobal,
    feedbacks,
    csv
)
{
    for (var i = 0; i < xmlIndices.length; ++i) {
        var xmlIndex = xmlIndices[i];
        var xmlDatapoints = xmlIndex.getElementsByTagName("datapointtypes")[0].childNodes;
        if (xmlDatapoints == undefined) {
            console.log("datapointtypes node not found for index:" + xmlIndex.getAttribute("index"));
            continue;
        }

        if (xmlDatapoints.length == 0) {
            console.log("datapointtypes length == 0 for index:" + xmlIndex.getAttribute("index"));
            continue;
        }

        var strAddressScenario = undefined;
        var strIndex = xmlIndex.getAttribute("index");
        var strLabel = xmlIndex.getAttribute("label");
        var strRoom = xmlIndex.getAttribute("room");
        var strZone = xmlIndex.getAttribute("zone");

        // Collect names
        for (var d = 0; d < xmlDatapoints.length; ++d) {
            if (xmlDatapoints[d].childNodes.length > 0) {
                var strName = xmlDatapoints[d].getAttribute("name");
                if (strName == "DPTVimarScene") {
                    var xmlGroupObjects = xmlDatapoints[d].getElementsByTagName("groupobject");
                    if (xmlGroupObjects != undefined) {
                        for (var g = 0; g < xmlGroupObjects.length; ++g) {
                            strAddressScenario = xmlGroupObjects[g].getAttribute("address");
                            scenariesName[strAddressScenario] = strIndex + "," + strZone + "," + strRoom + "," + strLabel;
                            scenariesType[strAddressScenario] = bGlobal;
                        }
                    }
                }
            }
        }

        // Collect object addresses
        var xmlDevices = xmlIndex.getElementsByTagName("device");
        if (xmlDevices != undefined) {
            for (var d = 0; d < xmlDevices.length; ++d){
                var strHexAddress = xmlDevices[d].getAttribute("address");
                if (scenaries[strAddressScenario] == undefined) {
                    scenaries[strAddressScenario] = new Array();
                }
                scenaries[strAddressScenario].push(strHexAddress);
            }
        }
    }
}

function buildAutomation
(
    visitor,
    xmlIndices,
    gateways,
    scenaries,
    feedbacks,
    csv
)
{
    var groups = {
        automation: undefined,
        switches: undefined,
        sliders: undefined,
        shutters: undefined
    };

    for (var i = 0; i < xmlIndices.length; ++i) {
        var xmlIndex = xmlIndices[i];
        var xmlDatapoints = xmlIndex.getElementsByTagName("datapointtypes")[0].childNodes;
        if (xmlDatapoints == undefined) {
            console.log("datapointtypes node not found for index:" + xmlIndex.getAttribute("index"));
            continue;
        }

        if (xmlDatapoints.length == 0) {
            console.log("datapointtypes length == 0 for index:" + xmlIndex.getAttribute("index"));
            continue;
        }
        
        var xmlDevices = xmlIndex.getElementsByTagName("device");
        if (xmlDevices == undefined) {
            console.log("device node not found for index:" + xmlIndex.getAttribute("index"));
            continue;
        }

        if (xmlDevices.length == 0) {
            console.log("device length == 0 for index:" + xmlIndex.getAttribute("index"));
            continue;
        }

        var xmlDevice = undefined;
        var xmlFunctionalBlock = undefined;
        for (var d = 0; d < xmlDevices.length; ++d){
            xmlDevice = xmlDevices[d];
            var main = undefined;
            var xmlFunctionalBlocks = xmlDevice.getElementsByTagName("functionalblock");
            for (var f = 0; f < xmlFunctionalBlocks.length; ++f){
                xmlFunctionalBlock = xmlFunctionalBlocks[f];
                main = xmlFunctionalBlock.getAttribute("main");
                if (main == undefined) {
                    xmlFunctionalBlock = undefined;
                    continue;
                }

                if (main == "True") {
                    break;
                }
            }

            if (main == undefined) {
                xmlDevice = undefined;
                xmlFunctionalBlock = undefined;
                continue;
            }

            if (main == "True") {
                break;
            }
        }

        if (xmlDevice == undefined || xmlFunctionalBlock == undefined) {
            continue;
        }

        if (groups.automation == undefined) {
            groups.automation = application.addComponent({
                type: "group",
                name: "Vimar-" + _D("Automation")
            });
        }

        blr.executionStack.push(
            function (in_args, out_args) {
                out_args = create_automation_component(
                    in_args.visitor,
                    in_args.xmlIndex,
                    in_args.xmlDatapoints,
                    in_args.xmlDevice,
                    in_args.xmlFunctionalBlock,
                    in_args.groups,
                    in_args.gateways,
                    in_args.scenaries,
                    in_args.feedbacks,
                    in_args.csv
                );
            },
            {
                visitor: visitor,
                xmlIndex:xmlIndex,
                xmlDatapoints:xmlDatapoints,
                xmlDevice:xmlDevice,
                xmlFunctionalBlock: xmlFunctionalBlock,
                groups: groups,
                gateways: gateways,
                scenaries: scenaries,
                feedbacks:feedbacks,
                csv: csv
            }
        );
    }

    if (groups.automation != undefined) {
        blr.executionStack.push(
            function (in_args, out_args) {
                //automation: undefined,
                //switches: undefined,
                //sliders: undefined,
                //shutters: undefined
                if (in_args.groups.shutters != undefined) {
                    if (in_args.groups.shutters.children.length == 0) {
                        application.removeComponent(in_args.groups.shutters);
                    }
                }
                if (in_args.groups.sliders != undefined) {
                    if (in_args.groups.sliders.children.length == 0) {
                        application.removeComponent(in_args.groups.sliders);
                    }
                }
                if (in_args.groups.switches != undefined) {
                    if (in_args.groups.switches.children.length == 0) {
                        application.removeComponent(in_args.groups.switches);
                    }
                }
                if (in_args.groups.automation != undefined) {
                    if(in_args.groups.automation.children.length == 0){
                        application.removeComponent(in_args.groups.automation);
                    }
                }
            },
            {
                groups: groups
            }
        );
    }
}


function create_scenario_component
(
    visitor,
    groupScenaries,
    gateways,
    name,
    address,
    value,
    addresses,
    feedbacks,
    csv
)
{
    var oItem = undefined;
    try {
        oItem = visitor.getComponentBy(
            "vimar",
            "scenario",
            {
                "vimar/group-address": address,
                "vimar/value": value,
                "vimar/physical-addresses": addresses
            }
        );
        if (oItem == undefined) {
            oItem = application.addComponent({
                type: "scenario",
                gateway: gateways.vimar,
                name: name,
                "vimar/group-address": address,
                "vimar/value": value,
                "vimar/physical-addresses": addresses
            }, groupScenaries);
        }

        csv.data += build_csv_row(oItem.getData().getParamVal("name"), "17.001", vimarGroupToKnxGroup(address), "", value);

    } catch (err) {
        blr.executionStack.cancel();
        console.error(NAME + " error\n" + err);
        alert(NAME + " error\n" + err);
    }
    return oItem;
}

function create_automation_component
(
    visitor,
    xmlIndex,
    xmlDatapoints,
    xmlDevice,
    xmlFunctionalBlock,
    groups,
    gateways,
    scenaries,
    feedbacks,
    csv
)
{
    var oItem = undefined;
    try{
        var strAddressesScenario = new Array();
        var strIndex = xmlIndex.getAttribute("index");
        var strLabel = xmlIndex.getAttribute("label");
        var strRoom = xmlIndex.getAttribute("room");
        var strZone = xmlIndex.getAttribute("zone");
        var strHexAddress = xmlDevice.getAttribute("address");
        var hexType = parseInt(xmlDevice.getAttribute("type").toString(16), 16);
        var iFid = parseInt(xmlFunctionalBlock.getAttribute("index"), 10);
        
        // Update local scenaries
        var xmlFbGroupObjects = xmlFunctionalBlock.getElementsByTagName("groupobjects")[0];
        if (xmlFbGroupObjects != undefined) {
            var xmlFbDataType = xmlFbGroupObjects.getElementsByTagName("datapointtype");
            if (xmlFbDataType != undefined) {
                for (var i = 0; i < xmlFbDataType.length; ++i) {
                    var xmlDptType = xmlFbDataType[i];
                    var strName = xmlDptType.getAttribute("name");
                    if (strName == "DPTVimarScene") {
                        var xmlGroupObjects = xmlDptType.getElementsByTagName("groupobject");
                        if (xmlGroupObjects != undefined) {
                            for (var g = 0; g < xmlGroupObjects.length; ++g) {
                                var xmlGO = xmlGroupObjects[g];
                                var strAddressScenario = xmlGO.getAttribute("address");
                                
                                if (scenaries[strAddressScenario] == undefined) {
                                    scenaries[strAddressScenario] = new Array();
                                }
                                scenaries[strAddressScenario].push(strHexAddress);
                                strAddressesScenario.push(strAddressScenario);
                            }
                        }
                    }
                }
            }
        }
        


        //console.log("[" + strZone + "," + strRoom + "," + strIndex + "," + strLabel + "][" + hexType.toString(16) + "," + strHexAddress + "," + iFid + "]");
        switch (hexType) {
            // By-Me Push button
            case 0x60:
            case 0x64: {
                // Use for trigger only
                if (groups.switches == undefined) {
                    groups.switches = application.addComponent({
                        type: "group",
                        name: _D("Switches")
                    }, groups.automation);
                }

                // Just grab group address
                var aGroupAddresses = new Array();
                for (var i = 0; i < xmlDatapoints.length; ++i) {
                    if (xmlDatapoints[i].childNodes.length > 0) {
                        // DPTSwitch
                        var strType = xmlDatapoints[i].getAttribute("name");
                        switch (strType) {
                            case "DPTSwitch": {
                                var xmlGroupObjects = xmlDatapoints[i].getElementsByTagName("groupobject");
                                if (xmlGroupObjects != undefined) {
                                    for (var g = 0; g < xmlGroupObjects.length; ++g) {
                                        aGroupAddresses.push(xmlGroupObjects[g].getAttribute("address"));
                                    }
                                }
                            } break;
                        }
                    }
                }

                for (var g = 0; g < aGroupAddresses.length; ++g) {
                    oItem = visitor.getComponentBy(
                        "knx",
                        "switch",
                        {
                            savelastvalue: true,
                            cache: true,
                            "knx/write-address": vimarGroupToKnxGroup(aGroupAddresses[g]),
                            "knx/feedback-address": vimarGroupToKnxGroup(aGroupAddresses[g])
                        }
                    );
                    if (oItem == undefined) {
                        oItem = application.addComponent({
                            type: "switch",
                            gateway: gateways.knx,
                            name: strIndex + "(TR)," + strZone + "," + strRoom + "," + strLabel,
                            savelastvalue: true,
                            cache: true,
                            "knx/write-address": vimarGroupToKnxGroup(aGroupAddresses[g]),
                            "knx/feedback-address": vimarGroupToKnxGroup(aGroupAddresses[g]),
                            "knx/virtual-feedback": false,
                            "icons": [
                                blr.gfxResource.getIconByTypeName(0, "off"),
                                blr.gfxResource.getIconByTypeName(1, "on")
                            ]
                        }, groups.switches);
                    }
                }

                /*
                // We do not need a feedback here...
                oItem = visitor.getComponentBy(
                    "knx",
                    "switch",
                    {
                        "knx/write-address": vimarGroupToKnxGroup(feedbacks.address),
                        "knx/feedback-address": vimarGroupToKnxGroup(feedbacks.address + 1),
                        "knx/virtual-feedback": true
                    }
                );
                if (oItem == undefined) {
                    oItem = application.addComponent({
                        type: "switch",
                        gateway: gateways.knx,
                        name: strIndex + "(VF)," + strZone + "," + strRoom + "," + strLabel,
                        "knx/write-address": vimarGroupToKnxGroup(feedbacks.address),
                        "knx/feedback-address": vimarGroupToKnxGroup(feedbacks.address + 1),
                        "knx/virtual-feedback": true,
                        "icons": [
                            blr.gfxResource.getIconByTypeName(0, "off"),
                            blr.gfxResource.getIconByTypeName(1, "on")
                        ]
                    }, groups.switches);
                }
                */

            } break;
            // By-Me Only switch
            case 0x65:
            case 0x61:
            case 0x85:
            case 0x82:
            case 0x69:
            case 0x91:
            case 0x6B:
            case 0x73:
            case 0x66:
            case 0x68: {
                if (groups.switches == undefined) {
                    groups.switches = application.addComponent({
                        type: "group",
                        name: _D("Switches")
                    },groups.automation);
                }

                // Find lower group addresses as component group address 
                var aGroupAddresses = new Array();
                for (var i = 0; i < xmlDatapoints.length; ++i) {
                    if (xmlDatapoints[i].childNodes.length > 0) {
                        var xmlGroupObjects = xmlDatapoints[i].getElementsByTagName("groupobject");
                        if (xmlGroupObjects != undefined) {
                            for (var g = 0; g < xmlGroupObjects.length; ++g) {
                                aGroupAddresses.push(xmlGroupObjects[g].getAttribute("address"));
                            }
                        }
                    }
                }
                var strGroupAddress = aGroupAddresses.sort()[0];

                oItem = visitor.getComponentBy(
                    "vimar",
                    "switch",
                    {
                        "vimar/group-address": strGroupAddress,
                        "vimar/physical-address": strHexAddress,
                        "vimar/device-pid": hexType.toString(16).toUpperCase(),
                        "vimar/device-fid": iFid
                    }
                );

                if (oItem == undefined) {
                    oItem = application.addComponent({
                        type: "switch",
                        gateway: gateways.vimar,
                        name: strIndex + "," + strZone + "," + strRoom + "," + strLabel,
                        "vimar/group-address": strGroupAddress,
                        "vimar/physical-address": strHexAddress,
                        "vimar/device-pid": hexType.toString(16).toUpperCase(),
                        "vimar/device-fid": iFid,
                        "icons": [
                            blr.gfxResource.getIconByTypeName(0, "off"),
                            blr.gfxResource.getIconByTypeName(1, "on")
                        ]
                    }, groups.switches);
                }

                if (feedbacks.create) {
                    oItem = visitor.getComponentBy(
                        "knx",
                        "switch",
                        {
                            "knx/write-address": vimarGroupToKnxGroup(feedbacks.address),
                            "knx/feedback-address": vimarGroupToKnxGroup(feedbacks.address + 1),
                            "knx/virtual-feedback":true
                        }
                    );
                    if (oItem == undefined) {
                        oItem = application.addComponent({
                            type: "switch",
                            gateway: gateways.knx,
                            name: strIndex + "(VF)," + strZone + "," + strRoom + "," + strLabel,
                            "knx/write-address": vimarGroupToKnxGroup(feedbacks.address),
                            "knx/feedback-address": vimarGroupToKnxGroup(feedbacks.address + 1),
                            "knx/virtual-feedback":true,
                            "icons": [
                                blr.gfxResource.getIconByTypeName(0, "off"),
                                blr.gfxResource.getIconByTypeName(1, "on")
                            ]
                        }, groups.switches);
                    }

                    csv.data += build_csv_row(oItem.getData().getParamVal("name"), "1.001", vimarGroupToKnxGroup(feedbacks.address), vimarGroupToKnxGroup(feedbacks.address+1), "");
                    feedbacks.address += 2;
                } else {
                    csv.data += build_csv_row(oItem.getData().getParamVal("name"), "1.001", vimarGroupToKnxGroup(strGroupAddress), "", "");
                }
            } break;
            // By-Me Switch & Sliders
            case 0x63:
            case 0x87:
            case 0x6F:
            case 0x7B:
            case 0xA1:
            case 0xA2:
            case 0xA3: {
                if (groups.switches == undefined) {
                    groups.switches = application.addComponent({
                        type: "group",
                        name: _D("Switches")
                    }, groups.automation);
                }
                if (groups.sliders == undefined) {
                    groups.sliders = application.addComponent({
                        type: "group",
                        name: _D("Sliders")
                    }, groups.automation);
                }
                // Find lower group addresses as component group address 
                var aGroupAddresses = new Array();
                for (var i = 0; i < xmlDatapoints.length; ++i) {
                    if (xmlDatapoints[i].childNodes.length > 0) {
                        var strName = xmlDatapoints[i].getAttribute("name");
                        var xmlGroupObjects = xmlDatapoints[i].getElementsByTagName("groupobject");
                        if (xmlGroupObjects != undefined) {
                            for (var g = 0; g < xmlGroupObjects.length; ++g) {
                                var strAddress = xmlGroupObjects[g].getAttribute("address");
                                aGroupAddresses.push(strAddress);
                            }
                        }
                    }
                }
                var strGroupAddress = aGroupAddresses.sort()[0];

                oItem = visitor.getComponentBy(
                    "vimar",
                    "switch",
                    {
                        "vimar/group-address": strGroupAddress,
                        "vimar/physical-address": strHexAddress,
                        "vimar/device-pid": hexType.toString(16).toUpperCase(),
                        "vimar/device-fid": iFid
                    }
                );
                if (oItem == undefined) {
                    oItem = application.addComponent({
                        type: "switch",
                        gateway: gateways.vimar,
                        name: strIndex + "," + strZone + "," + strRoom + "," + strLabel,
                        "vimar/group-address": strGroupAddress,
                        "vimar/physical-address": strHexAddress,
                        "vimar/device-pid": hexType.toString(16).toUpperCase(),
                        "vimar/device-fid": iFid,
                        "icons": [
                            blr.gfxResource.getIconByTypeName(0, "off"),
                            blr.gfxResource.getIconByTypeName(1, "on")
                        ]
                    }, groups.switches);
                }

                oItem = visitor.getComponentBy(
                    "vimar",
                    "slider",
                    {
                        "vimar/group-address": strGroupAddress,
                        "vimar/physical-address": strHexAddress,
                        "vimar/device-pid": hexType.toString(16).toUpperCase(),
                        "vimar/device-fid": iFid
                    }
                );
                if (oItem == undefined) {
                    oItem = application.addComponent({
                        type: "slider",
                        gateway: gateways.vimar,
                        name: strIndex + "," + strZone + "," + strRoom + "," + strLabel,
                        "vimar/group-address": strGroupAddress,
                        "vimar/physical-address": strHexAddress,
                        "vimar/device-pid": hexType.toString(16).toUpperCase(),
                        "vimar/device-fid": iFid,
                        "icons": [
                            blr.gfxResource.getIconByTypeName(0, "volume")
                        ]
                    }, groups.sliders);
                }

                if (feedbacks.create) {
                    oItem = visitor.getComponentBy(
                        "knx",
                        "switch",
                        {
                            "knx/write-address": vimarGroupToKnxGroup(feedbacks.address),
                            "knx/feedback-address": vimarGroupToKnxGroup(feedbacks.address + 1),
                            "knx/virtual-feedback": true
                        }
                    );
                    if (oItem == undefined) {
                        oItem = application.addComponent({
                            type: "switch",
                            gateway: gateways.knx,
                            name: strIndex + "(VF)," + strZone + "," + strRoom + "," + strLabel,
                            "knx/write-address": vimarGroupToKnxGroup(feedbacks.address),
                            "knx/feedback-address": vimarGroupToKnxGroup(feedbacks.address + 1),
                            "knx/virtual-feedback": true,
                            "icons": [
                                blr.gfxResource.getIconByTypeName(0, "off"),
                                blr.gfxResource.getIconByTypeName(1, "on")
                            ]
                        }, groups.switches);
                    }
                    csv.data += build_csv_row(oItem.getData().getParamVal("name"), "3.007", vimarGroupToKnxGroup(parseInt(strGroupAddress, 16) + 1), "", "");

                    csv.data += build_csv_row(oItem.getData().getParamVal("name"), "1.001", vimarGroupToKnxGroup(feedbacks.address), vimarGroupToKnxGroup(feedbacks.address + 1), "");
                    feedbacks.address += 2;
                    
                    oItem = visitor.getComponentBy(
                        "knx",
                        "slider",
                        {
                            "knx/write-address": vimarGroupToKnxGroup(feedbacks.address),
                            "knx/feedback-address": vimarGroupToKnxGroup(feedbacks.address + 1),
                            "knx/virtual-feedback": true
                        }
                    );
                    if (oItem == undefined) {
                        oItem = application.addComponent({
                            type: "slider",
                            gateway: gateways.knx,
                            name: strIndex + "(VF)," + strZone + "," + strRoom + "," + strLabel,
                            "knx/write-address": vimarGroupToKnxGroup(feedbacks.address),
                            "knx/feedback-address": vimarGroupToKnxGroup(feedbacks.address + 1),
                            "knx/virtual-feedback": true,
                            "icons": [
                                blr.gfxResource.getIconByTypeName(0, "volume")
                            ]
                        }, groups.sliders);
                    }

                    csv.data += build_csv_row(oItem.getData().getParamVal("name"), "5.001", vimarGroupToKnxGroup(feedbacks.address), vimarGroupToKnxGroup(feedbacks.address + 1), "");
                    feedbacks.address += 2;

                } else {
                    csv.data += build_csv_row(oItem.getData().getParamVal("name"), "1.001", vimarGroupToKnxGroup(strGroupAddress), "", "");
                    csv.data += build_csv_row(oItem.getData().getParamVal("name"), "5.001", vimarGroupToKnxGroup(strGroupAddress), "", "");
                    csv.data += build_csv_row(oItem.getData().getParamVal("name"), "3.007", vimarGroupToKnxGroup(parseInt(strGroupAddress, 16) + 1), "", "");
                }
            } break;
            // Knx Switch & sliders
            case 0xAA: {
                if (groups.switches == undefined) {
                    groups.switches = application.addComponent({
                        type: "group",
                        name: _D("Switches")
                    }, groups.automation);
                }
                if (groups.sliders == undefined) {
                    groups.sliders = application.addComponent({
                        type: "group",
                        name: _D("Sliders")
                    }, groups.automation);
                }

                var switchWriteAdr = 0;
                var switchReadAdr = 0;
                var brightWriteAdr = 0;
                var brightReadAdr = 0;
                var dimmingWriteAdr = 0;
                for (var i = 0; i < xmlDatapoints.length; ++i) {
                    if (xmlDatapoints[i].childNodes.length > 0) {
                        var strName = xmlDatapoints[i].getAttribute("name");
                        var xmlGroupObjects = xmlDatapoints[i].getElementsByTagName("groupobject");
                        switch(strName){
                            case "DPTSwitch": {
                                switchWriteAdr = parseInt(xmlGroupObjects[0].getAttribute("address"), 16);
                            } break;
                            case "DPT_Info_Switch": {
                                switchReadAdr = parseInt(xmlGroupObjects[0].getAttribute("address"), 16);
                            } break;
                            case "DPT_Brightness": {
                                brightWriteAdr = parseInt(xmlGroupObjects[0].getAttribute("address"), 16);
                            } break;
                            case "DPT_Info_Brightness": {
                                brightReadAdr = parseInt(xmlGroupObjects[0].getAttribute("address"), 16);
                            } break;
                            case "DPTControlDimming": {
                                dimmingWriteAdr = parseInt(xmlGroupObjects[0].getAttribute("address"), 16);
                            } break;
                        }
                    }
                }

                if (switchWriteAdr != 0 && switchReadAdr != 0) {
                    oItem = visitor.getComponentBy(
                        "knx",
                        "switch",
                        {
                            "knx/write-address": vimarGroupToKnxGroup(switchWriteAdr),
                            "knx/feedback-address": vimarGroupToKnxGroup(switchReadAdr)
                        }
                    );
                    if (oItem == undefined) {
                        oItem = application.addComponent({
                            type: "switch",
                            gateway: gateways.knx,
                            name: strIndex + "," + strZone + "," + strRoom + "," + strLabel + " " + _D("On-Off"),
                            "knx/write-address": vimarGroupToKnxGroup(switchWriteAdr),
                            "knx/feedback-address": vimarGroupToKnxGroup(switchReadAdr),
                            "icons": [
                                blr.gfxResource.getIconByTypeName(0, "off"),
                                blr.gfxResource.getIconByTypeName(1, "on")
                            ]
                        }, groups.switches);
                    }


                    csv.data += build_csv_row(
                        oItem.getData().getParamVal("name"),
                        "1.001",
                        vimarGroupToKnxGroup(switchWriteAdr),
                        vimarGroupToKnxGroup(switchReadAdr),
                        ""
                    );
                }

                if (brightWriteAdr != 0 && brightReadAdr != 0) {
                    oItem = visitor.getComponentBy(
                        "knx",
                        "slider",
                        {
                            "knx/write-address": vimarGroupToKnxGroup(brightWriteAdr),
                            "knx/feedback-address": vimarGroupToKnxGroup(brightReadAdr)
                        }
                    );
                    if (oItem == undefined) {
                        oItem = application.addComponent({
                            type: "slider",
                            gateway: gateways.knx,
                            name: strIndex + "," + strZone + "," + strRoom + "," + strLabel + " " + _D("Intensity"),
                            "knx/write-address": vimarGroupToKnxGroup(brightWriteAdr),
                            "knx/feedback-address": vimarGroupToKnxGroup(brightReadAdr),
                            "icons": [
                                blr.gfxResource.getIconByTypeName(0, "volume")
                            ]
                        }, groups.sliders);
                    }

                    csv.data += build_csv_row(
                        oItem.getData().getParamVal("name"),
                        "5.001",
                        vimarGroupToKnxGroup(brightWriteAdr),
                        vimarGroupToKnxGroup(brightReadAdr),
                        ""
                    );

                    if (dimmingWriteAdr != 0) {
                        csv.data += build_csv_row(
                            oItem.getData().getParamVal("name"),
                            "3.007",
                            vimarGroupToKnxGroup(dimmingWriteAdr),
                            "",
                            ""
                        );
                    }
                }
                
                if (strAddressesScenario.length > 0) {// this object is inside a scenario
                    // We need to add the switchReadAdr and brightReadAdr addresses
                    for (var sa = 0; sa < strAddressesScenario.length; ++sa) {
                        var aData = scenaries[strAddressesScenario[sa]];
                        aData.pop();
                        // We just added
                        if (switchReadAdr != 0) { aData.push("G" + switchReadAdr); }
                        if (brightReadAdr != 0) { aData.push("G" + brightReadAdr); }
                    }
                }
            } break;

            // RGB Light
            // case 0xA9: { } break;

            // By-Me Shutter
            case 0x62:
            case 0x86:
            case 0x75: {
                if (groups.shutters == undefined) {
                    groups.shutters = application.addComponent({
                        type: "group",
                        name: _D("Shutters")
                    }, groups.automation);
                }

                // Find lower group addresses as component group address 
                var aGroupAddresses = new Array();
                for (var i = 0; i < xmlDatapoints.length; ++i) {
                    if (xmlDatapoints[i].childNodes.length > 0) {
                        var xmlGroupObjects = xmlDatapoints[i].getElementsByTagName("groupobject");
                        if (xmlGroupObjects != undefined) {
                            for (var g = 0; g < xmlGroupObjects.length; ++g) {
                                aGroupAddresses.push(xmlGroupObjects[g].getAttribute("address"));
                            }
                        }
                    }
                }
                var strGroupAddress = aGroupAddresses.sort()[0];

                oItem = visitor.getComponentBy(
                    "vimar",
                    "shutter",
                    {
                        "vimar/group-address": strGroupAddress,
                        "vimar/physical-address": strHexAddress,
                        "vimar/device-pid": hexType.toString(16).toUpperCase(),
                        "vimar/device-fid": iFid
                    }
                );
                if (oItem == undefined) {
                    oItem = application.addComponent({
                        type: "shutter",
                        gateway: gateways.vimar,
                        name: strIndex + "," + strZone + "," + strRoom + "," + strLabel,
                        "vimar/group-address": strGroupAddress,
                        "vimar/physical-address": strHexAddress,
                        "vimar/device-pid": hexType.toString(16).toUpperCase(),
                        "vimar/device-fid": iFid
                    }, groups.shutters);
                }

                csv.data += build_csv_row(oItem.getData().getParamVal("name"), "1.007", vimarGroupToKnxGroup(strGroupAddress), "", "");
                csv.data += build_csv_row(oItem.getData().getParamVal("name"), "1.008", vimarGroupToKnxGroup(parseInt(strGroupAddress, 16) + 1), "", "");

            } break;
            default: {
                console.warn("Unknown automation device [" + hexType.toString(16).toUpperCase() + "]" + strIndex + "," + strZone + "," + strRoom + "," + strLabel);
            } break;
        }
    } catch (err) {
        blr.executionStack.cancel();
        console.error(NAME + " error\n" + err);
        alert(NAME + " error\n" + err);
    }
    return oItem;
}

function create_power_component
(
    visitor,
    xmlIndex,
    xmlDatapoints,
    xmlDevice,
    xmlFunctionalBlock,
    groups,
    gateways,
    scenaries,
    feedbacks,
    csv
)
{
    var oItem = undefined;
    try {
        var strAddressesScenario = new Array();
        var strIndex = xmlIndex.getAttribute("index");
        var strLabel = xmlIndex.getAttribute("label");
        var strRoom = xmlIndex.getAttribute("room");
        var strZone = xmlIndex.getAttribute("zone");
        var strHexAddress = xmlDevice.getAttribute("address");
        var hexType = parseInt(xmlDevice.getAttribute("type").toString(16), 16);
        var iFid = parseInt(xmlFunctionalBlock.getAttribute("index"), 10);

        // Update local scenaries
        var xmlFbGroupObjects = xmlFunctionalBlock.getElementsByTagName("groupobjects")[0];
        if (xmlFbGroupObjects != undefined) {
            var xmlFbDataType = xmlFbGroupObjects.getElementsByTagName("datapointtype");
            if (xmlFbDataType != undefined) {
                for (var i = 0; i < xmlFbDataType.length; ++i) {
                    var xmlDptType = xmlFbDataType[i];
                    var strName = xmlDptType.getAttribute("name");
                    if (strName == "DPTVimarScene") {
                        var xmlGroupObjects = xmlDptType.getElementsByTagName("groupobject");
                        if (xmlGroupObjects != undefined) {
                            for (var g = 0; g < xmlGroupObjects.length; ++g) {
                                var xmlGO = xmlGroupObjects[g];
                                var strAddressScenario = xmlGO.getAttribute("address");

                                if (scenaries[strAddressScenario] == undefined) {
                                    scenaries[strAddressScenario] = new Array();
                                }
                                scenaries[strAddressScenario].push(strHexAddress);
                                strAddressesScenario.push(strAddressScenario);
                            }
                        }
                    }
                }
            }
        }

        switch (hexType) {
            // By-Me Load Control
            case 0x6E: {
                // Find lower group addresses as component group address 
                var aGroupAddresses = new Array();
                for (var i = 0; i < xmlDatapoints.length; ++i) {
                    if (xmlDatapoints[i].childNodes.length > 0) {
                        var xmlGroupObjects = xmlDatapoints[i].getElementsByTagName("groupobject");
                        if (xmlGroupObjects != undefined) {
                            for (var g = 0; g < xmlGroupObjects.length; ++g) {
                                aGroupAddresses.push(xmlGroupObjects[g].getAttribute("address"));
                            }
                        }
                    }
                }
                var strGroupAddress = aGroupAddresses.sort()[0];

                oItem = visitor.getComponentBy(
                    "vimar",
                    "info",
                    {
                        "vimar/group-address": strGroupAddress,
                        "vimar/physical-address": strHexAddress,
                        "vimar/device-pid": hexType.toString(16).toUpperCase(),
                        "vimar/device-fid": iFid
                    }
                );
                if (oItem == undefined) {
                    oItem = application.addComponent({
                        type: "info",
                        gateway: gateways.vimar,
                        name: strIndex + "," + strZone + "," + strRoom + "," + strLabel + " " + _D("Power"),
                        "vimar/group-address": strGroupAddress,
                        "vimar/physical-address": strHexAddress,
                        "vimar/device-pid": hexType.toString(16).toUpperCase(),
                        "vimar/device-fid": iFid
                    }, groups.power);
                }
            } break;
            // Knx Load Control
            case 0xB0:
            case 0xB1: {
                var activePowerAdr = 0;
                var reactivePowerAdr = 0;
                var absoluteEnergyAdr = 0;
                var partialEnergyAdr = 0;
                for (var i = 0; i < xmlDatapoints.length; ++i) {
                    if (xmlDatapoints[i].childNodes.length > 0) {
                        var strName = xmlDatapoints[i].getAttribute("name");
                        var xmlGroupObjects = xmlDatapoints[i].getElementsByTagName("groupobject");
                        switch (strName) {
                            case "DPT_PotenzaAttiva": {
                                activePowerAdr = parseInt(xmlGroupObjects[0].getAttribute("address"), 16);
                            } break;
                            case "DPT_PotenzaReattiva": {
                                reactivePowerAdr = parseInt(xmlGroupObjects[0].getAttribute("address"), 16);
                            } break;
                            case "DPT_Energia_Assoluta": {
                                absoluteEnergyAdr = parseInt(xmlGroupObjects[0].getAttribute("address"), 16);
                            } break;
                            case "DPT_Energia_Parziale": {
                                partialEnergyAdr = parseInt(xmlGroupObjects[0].getAttribute("address"), 16);
                            } break;
                        }
                    }
                }

                if (activePowerAdr != 0) {
                    oItem = visitor.getComponentBy(
                        "knx",
                        "info",
                        {
                            "knx/feedback-address" : vimarGroupToKnxGroup(activePowerAdr),
                            "knx/DPT": "14.56",
                            "knx/raw-value":true
                        }
                    );
                    if (oItem == undefined) {
                        oItem = application.addComponent({
                            type: "info",
                            gateway: gateways.knx,
                            name: strIndex + "," + strZone + "," + strRoom + "," + strLabel + " " + _D("Active Power"),
                            "knx/feedback-address": vimarGroupToKnxGroup(activePowerAdr),
                            "knx/DPT": "14.56",
                            "knx/raw-value": true
                        }, groups.power);
                    }

                    csv.data += build_csv_row(
                        oItem.getData().getParamVal("name"),
                        "14.056",
                        "",
                        vimarGroupToKnxGroup(activePowerAdr),
                        ""
                    );
                }

                if (reactivePowerAdr != 0) {
                    oItem = visitor.getComponentBy(
                        "knx",
                        "info",
                        {
                            "knx/feedback-address": vimarGroupToKnxGroup(reactivePowerAdr),
                            "knx/DPT": "14.56",
                            "knx/raw-value": true
                        }
                    );
                    if (oItem == undefined) {
                        oItem = application.addComponent({
                            type: "info",
                            gateway: gateways.knx,
                            name: strIndex + "," + strZone + "," + strRoom + "," + strLabel + " " + _D("Reactive Power"),
                            "knx/feedback-address": vimarGroupToKnxGroup(reactivePowerAdr),
                            "knx/DPT": "14.56",
                            "knx/raw-value": true
                        }, groups.power);
                    }

                    csv.data += build_csv_row(
                        oItem.getData().getParamVal("name"),
                        "14.056",
                        "",
                        vimarGroupToKnxGroup(reactivePowerAdr),
                        ""
                    );
                }

                if (absoluteEnergyAdr != 0) {
                    oItem = visitor.getComponentBy(
                        "knx",
                        "info",
                        {
                            "knx/feedback-address": vimarGroupToKnxGroup(absoluteEnergyAdr),
                            "knx/DPT": "12.1",
                            "knx/raw-value": true
                        }
                    );
                    if (oItem == undefined) {
                        oItem = application.addComponent({
                            type: "info",
                            gateway: gateways.knx,
                            name: strIndex + "," + strZone + "," + strRoom + "," + strLabel + " " + _D("Absolute Energy"),
                            "knx/feedback-address": vimarGroupToKnxGroup(absoluteEnergyAdr),
                            "knx/DPT": "12.1",
                            "knx/raw-value": true
                        }, groups.power);
                    }

                    csv.data += build_csv_row(
                        oItem.getData().getParamVal("name"),
                        "12.001",
                        "",
                        vimarGroupToKnxGroup(absoluteEnergyAdr),
                        ""
                    );
                }

                if (partialEnergyAdr != 0) {
                    oItem = visitor.getComponentBy(
                        "knx",
                        "info",
                        {
                            "knx/feedback-address": vimarGroupToKnxGroup(partialEnergyAdr),
                            "knx/DPT": "12.1",
                            "knx/raw-value": true
                        }
                    );
                    if (oItem == undefined) {
                        oItem = application.addComponent({
                            type: "info",
                            gateway: gateways.knx,
                            name: strIndex + "," + strZone + "," + strRoom + "," + strLabel + " " + _D("Partial Energy"),
                            "knx/feedback-address": vimarGroupToKnxGroup(partialEnergyAdr),
                            "knx/DPT": "12.1",
                            "knx/raw-value": true
                        }, groups.power);
                    }

                    csv.data += build_csv_row(
                        oItem.getData().getParamVal("name"),
                        "12.001",
                        "",
                        vimarGroupToKnxGroup(partialEnergyAdr),
                        ""
                    );
                }

                if (strAddressesScenario.length > 0) {// this object is inside a scenario
                    // We need to add the switchReadAdr and brightReadAdr addresses
                    for (var sa = 0; sa < strAddressesScenario.length; ++sa) {
                        var aData = scenaries[strAddressesScenario[sa]];
                        aData.pop();
                        // We just added
                        if (activePowerAdr != 0) { aData.push("G" + activePowerAdr); }
                        if (reactivePowerAdr != 0) { aData.push("G" + reactivePowerAdr); }
                        if (absoluteEnergyAdr != 0) { aData.push("G" + absoluteEnergyAdr); }
                        if (partialEnergyAdr != 0) { aData.push("G" + partialEnergyAdr); }
                    }
                }

            } break;
            default: {
                console.warn("Unknown power device [" + hexType.toString(16).toUpperCase() + "]" + strIndex + "," + strZone + "," + strRoom + "," + strLabel);
            } break;
        }
    } catch (err) {
        blr.executionStack.cancel();
        console.error(NAME + " error\n" + err);
        alert(NAME + " error\n" + err);
    }
    return oItem;
}


function create_clima_component
(
    visitor,
    xmlIndex,
    xmlDatapoints,
    xmlDevice,
    xmlFunctionalBlock,
    groups,
    gateways,
    scenaries,
    feedbacks,
    csv
)
{
        var oItem = undefined;
        try {
            var strAddressesScenario = new Array();
            var strIndex = xmlIndex.getAttribute("index");
            var strLabel = xmlIndex.getAttribute("label");
            var strRoom = xmlIndex.getAttribute("room");
            var strZone = xmlIndex.getAttribute("zone");
            var strHexAddress = xmlDevice.getAttribute("address");
            var hexType = parseInt(xmlDevice.getAttribute("type").toString(16), 16);
            var iFid = parseInt(xmlFunctionalBlock.getAttribute("index"), 10);

            // Update local scenaries
            var xmlFbGroupObjects = xmlFunctionalBlock.getElementsByTagName("groupobjects")[0];
            if (xmlFbGroupObjects != undefined) {
                var xmlFbDataType = xmlFbGroupObjects.getElementsByTagName("datapointtype");
                if (xmlFbDataType != undefined) {
                    for (var i = 0; i < xmlFbDataType.length; ++i) {
                        var xmlDptType = xmlFbDataType[i];
                        var strName = xmlDptType.getAttribute("name");
                        if (strName == "DPTVimarScene") {
                            var xmlGroupObjects = xmlDptType.getElementsByTagName("groupobject");
                            if (xmlGroupObjects != undefined) {
                                for (var g = 0; g < xmlGroupObjects.length; ++g) {
                                    var xmlGO = xmlGroupObjects[g];
                                    var strAddressScenario = xmlGO.getAttribute("address");

                                    if (scenaries[strAddressScenario] == undefined) {
                                        scenaries[strAddressScenario] = new Array();
                                    }
                                    scenaries[strAddressScenario].push(strHexAddress);
                                    strAddressesScenario.push(strAddressScenario);
                                }
                            }
                        }
                    }
                }
            }

            switch (hexType) {
                // By-Me Valve Status
                case 0x66: {
                    // Find lower group addresses as component group address 
                    var aGroupAddresses = new Array();
                    for (var i = 0; i < xmlDatapoints.length; ++i) {
                        if (xmlDatapoints[i].childNodes.length > 0) {
                            var xmlGroupObjects = xmlDatapoints[i].getElementsByTagName("groupobject");
                            if (xmlGroupObjects != undefined) {
                                for (var g = 0; g < xmlGroupObjects.length; ++g) {
                                    aGroupAddresses.push(xmlGroupObjects[g].getAttribute("address"));
                                }
                            }
                        }
                    }
                    var strGroupAddress = aGroupAddresses.sort()[0];

                    oItem = visitor.getComponentBy(
                        "vimar",
                        "info-text",
                        {
                            "vimar/group-address": strGroupAddress,
                            "vimar/physical-address": strHexAddress,
                            "vimar/device-pid": hexType.toString(16).toUpperCase(),
                            "vimar/device-fid": iFid
                        }
                    );

                    if (oItem == undefined) {
                        oItem = application.addComponent({
                            type: "info-text",
                            gateway: gateways.vimar,
                            name: strIndex + "," + strZone + "," + strRoom + "," + strLabel + " " + _D("Valve"),
                            "vimar/group-address": strGroupAddress,
                            "vimar/physical-address": strHexAddress,
                            "vimar/device-pid": hexType.toString(16).toUpperCase(),
                            "vimar/device-fid": iFid,
                            "text-0": "Off",
                            "text-1": "On",
                            "color-0": "#333333",
                            "color-1": "#990000"
                        }, groups.clima);
                    }

                    if (feedbacks.create) {
                        oItem = visitor.getComponentBy(
                            "knx",
                            "info-text",
                            {
                                "knx/write-address": vimarGroupToKnxGroup(feedbacks.address),
                                "knx/feedback-address": vimarGroupToKnxGroup(feedbacks.address),
                                "knx/virtual-feedback": true
                            }
                        );
                        if (oItem == undefined) {
                            oItem = application.addComponent({
                                type: "info-text",
                                gateway: gateways.knx,
                                name: strIndex + "(VF)," + strZone + "," + strRoom + "," + strLabel + " " + _D("Valve"),
                                "knx/write-address": vimarGroupToKnxGroup(feedbacks.address),
                                "knx/feedback-address": vimarGroupToKnxGroup(feedbacks.address),
                                "knx/virtual-feedback": true,
                                "text-0": "Off",
                                "text-1": "On",
                                "color-0": "#333333",
                                "color-1": "#990000"
                            }, groups.clima);
                        }

                        csv.data += build_csv_row(oItem.getData().getParamVal("name"), "1.001", vimarGroupToKnxGroup(feedbacks.address), vimarGroupToKnxGroup(feedbacks.address), "");
                        feedbacks.address += 1;
                    } else {
                        csv.data += build_csv_row(oItem.getData().getParamVal("name"), "1.001", vimarGroupToKnxGroup(strGroupAddress), "", "");
                    }
                } break;
                // By-Me Thermostat
                case 0x6C: {
                    // Find lower group addresses as component group address 
                    var aGroupAddresses = new Array();
                    for (var i = 0; i < xmlDatapoints.length; ++i) {
                        if (xmlDatapoints[i].childNodes.length > 0) {
                            var xmlGroupObjects = xmlDatapoints[i].getElementsByTagName("groupobject");
                            if (xmlGroupObjects != undefined) {
                                for (var g = 0; g < xmlGroupObjects.length; ++g) {
                                    aGroupAddresses.push(xmlGroupObjects[g].getAttribute("address"));
                                }
                            }
                        }
                    }
                    var strGroupAddress = aGroupAddresses.sort()[0];
                    // Ambient Temperature
                    oItem = visitor.getComponentBy(
                        "vimar",
                        "info",
                        {
                            "vimar/group-address": strGroupAddress,
                            "vimar/physical-address": strHexAddress,
                            "vimar/device-pid": hexType.toString(16).toUpperCase(),
                            "vimar/device-fid": 8
                        }
                    );
                    if (oItem == undefined) {
                        oItem = application.addComponent({
                            type: "info",
                            gateway: gateways.vimar,
                            name: strIndex + "," + strZone + "," + strRoom + "," + strLabel + " " + _D("Ambient Temperature"),
                            "vimar/group-address": strGroupAddress,
                            "vimar/physical-address": strHexAddress,
                            "vimar/device-pid": hexType.toString(16).toUpperCase(),
                            "vimar/device-fid": 8
                        }, groups.clima);
                    }

                    if(feedbacks.create){
                        oItem = visitor.getComponentBy(
                            "knx",
                            "info",
                            {
                                "knx/write-address": vimarGroupToKnxGroup(feedbacks.address),
                                "knx/feedback-address": vimarGroupToKnxGroup(feedbacks.address),
                                "knx/virtual-feedback": true,
                                "knx/raw-value": true,
                                "knx/DPT": "9.1"
                            }
                        );
                        if (oItem == undefined) {
                            oItem = application.addComponent({
                                type: "info",
                                gateway: gateways.knx,
                                name: strIndex + "(VF)," + strZone + "," + strRoom + "," + strLabel + " " + _D("Ambient Temperature"),
                                "knx/write-address": vimarGroupToKnxGroup(feedbacks.address),
                                "knx/feedback-address": vimarGroupToKnxGroup(feedbacks.address),
                                "knx/virtual-feedback": true,
                                "knx/raw-value": true,
                                "knx/DPT":"9.1"
                            }, groups.clima);
                        }

                        csv.data += build_csv_row(oItem.getData().getParamVal("name"), "9.001", "", vimarGroupToKnxGroup(feedbacks.address), "");
                        feedbacks.address += 1;
                    }

                    // Set point
                    oItem = visitor.getComponentBy(
                        "vimar",
                        "thermostat",
                        {
                            "vimar/group-address": strGroupAddress,
                            "vimar/physical-address": strHexAddress,
                            "vimar/device-pid": hexType.toString(16).toUpperCase(),
                            "vimar/device-fid": 6
                        }
                    );
                    if (oItem == undefined) {
                        oItem = application.addComponent({
                            type: "thermostat",
                            gateway: gateways.vimar,
                            name: strIndex + "," + strZone + "," + strRoom + "," + strLabel,
                            "vimar/group-address": strGroupAddress,
                            "vimar/physical-address": strHexAddress,
                            "vimar/device-pid": hexType.toString(16).toUpperCase(),
                            "vimar/device-fid": 6,
                            "icons": [
                                blr.gfxResource.getIconByTypeName(0, "temperature"),
                                blr.gfxResource.getIconByTypeName(1, "less"),
                                blr.gfxResource.getIconByTypeName(2, "circle"),
                                blr.gfxResource.getIconByTypeName(3, "plus")
                            ]
                        }, groups.clima);
                    }

                    if (feedbacks.create) {
                        oItem = visitor.getComponentBy(
                            "knx",
                            "thermostat",
                            {
                                "knx/write-address": vimarGroupToKnxGroup(feedbacks.address),
                                "knx/feedback-address": vimarGroupToKnxGroup(feedbacks.address+1),
                                "knx/virtual-feedback": true
                            }
                        );
                        if (oItem == undefined) {
                            oItem = application.addComponent({
                                type: "thermostat",
                                gateway: gateways.knx,
                                name: strIndex + "(VF)," + strZone + "," + strRoom + "," + strLabel,
                                "knx/write-address": vimarGroupToKnxGroup(feedbacks.address),
                                "knx/feedback-address": vimarGroupToKnxGroup(feedbacks.address+1),
                                "knx/virtual-feedback": true
                            }, groups.clima);
                        }

                        csv.data += build_csv_row(oItem.getData().getParamVal("name"), "9.001", vimarGroupToKnxGroup(feedbacks.address), vimarGroupToKnxGroup(feedbacks.address+1), "");
                        feedbacks.address += 2;
                    }

                    // Crono
                    oItem = visitor.getComponentBy(
                        "vimar",
                        "timed-thermostat",
                        {
                            "vimar/group-address": strGroupAddress,
                            "vimar/physical-address": strHexAddress,
                            "vimar/device-pid": hexType.toString(16).toUpperCase(),
                            "vimar/device-fid": 6
                        }
                    );
                    if (oItem == undefined) {
                        oItem = application.addComponent({
                            type: "timed-thermostat",
                            gateway: gateways.vimar,
                            name: strIndex + "," + strZone + "," + strRoom + "," + strLabel + " " + _D("Crono"),
                            "vimar/group-address": strGroupAddress,
                            "vimar/physical-address": strHexAddress,
                            "vimar/device-pid": hexType.toString(16).toUpperCase(),
                            "vimar/device-fid": 6
                        }, groups.clima);
                    }

                    // Season
                    oItem = visitor.getComponentBy(
                        "vimar",
                        "combo",
                        {
                            "vimar/group-address": strGroupAddress,
                            "vimar/physical-address": strHexAddress,
                            "vimar/device-pid": hexType.toString(16).toUpperCase(),
                            "vimar/device-fid": 3,
                        }
                    );
                    if (oItem == undefined) {
                        oItem = application.addComponent({
                            type: "combo",
                            gateway: gateways.vimar,
                            name: strIndex + "," + strZone + "," + strRoom + "," + strLabel + " " + _D("Season"),
                            "vimar/group-address": strGroupAddress,
                            "vimar/physical-address": strHexAddress,
                            "vimar/device-pid": hexType.toString(16).toUpperCase(),
                            "vimar/device-fid": 3,
                            "values": [{
                                "value": 0,
                                "description": _D("Winter")
                            }, {
                                "value": 1,
                                "description": _D("Summer")
                            }],
                            "icons": [
                                blr.gfxResource.getIconByTypeName(0, "winter"),
                                blr.gfxResource.getIconByTypeName(1, "summer")
                            ]
                        }, groups.clima);
                    }

                    if (feedbacks.create) {
                        // To Be define...
                    }

                    // Mode
                    oItem = visitor.getComponentBy(
                        "vimar",
                        "combo",
                        {
                            "vimar/group-address": strGroupAddress,
                            "vimar/physical-address": strHexAddress,
                            "vimar/device-pid": hexType.toString(16).toUpperCase(),
                            "vimar/device-fid": 1,
                        }
                    );
                    if (oItem == undefined) {
                        oItem = application.addComponent({
                            type: "combo",
                            gateway: gateways.vimar,
                            name: strIndex + "," + strZone + "," + strRoom + "," + strLabel + " " + _D("Mode"),
                            "vimar/group-address": strGroupAddress,
                            "vimar/physical-address": strHexAddress,
                            "vimar/device-pid": hexType.toString(16).toUpperCase(),
                            "vimar/device-fid": 1,
                            "values": [{
                                "value": 0,
                                "description": _D("Off")
                            }, {
                                "value": 1,
                                "description": _D("Forced Off")
                            }, {
                                "value": 2,
                                "description": _D("Off Period")
                            }, {
                                "value": 3,
                                "description": _D("Antifreeze")
                            }, {
                                "value": 4,
                                "description": _D("Night")
                            }, {
                                "value": 5,
                                "description": _D("Night Period")
                            }, {
                                "value": 6,
                                "description": _D("Manual")
                            }, {
                                "value": 7,
                                "description": _D("Manual Period")
                            }, {
                                "value": 8,
                                "description": _D("Automatic")
                            }],
                            "icons": [
                                blr.gfxResource.getIconByTypeName(0, "off"),
                                blr.gfxResource.getIconByTypeName(1, "forced"),
                                blr.gfxResource.getIconByTypeName(2, "off"),
                                blr.gfxResource.getIconByTypeName(3, "antifreeze"),
                                blr.gfxResource.getIconByTypeName(4, "night"),
                                blr.gfxResource.getIconByTypeName(5, "night"),
                                blr.gfxResource.getIconByTypeName(6, "manual"),
                                blr.gfxResource.getIconByTypeName(7, "manual"),
                                blr.gfxResource.getIconByTypeName(8, "automatic")
                            ]
                        }, groups.clima);
                    }

                    if (feedbacks.create) {
                        // Apply standard compatibility
                        oItem.getData().setParamVal(
                            "values",
                            [
                                {
                                    "value": 8,
                                    "description": _D("Automatic")
                                },
                                {
                                    "value": 6,
                                    "description": _D("Manual")
                                },
                                {
                                    "value": 0,
                                    "description": _D("Off")
                                },
                                {
                                    "value": 4,
                                    "description": _D("Night")
                                },
                                {
                                    "value": 3,
                                    "description": _D("Antifreeze")
                                }
                            ]
                        );

                        oItem.getData().setParamVal(
                            "icons",
                            [
                                blr.gfxResource.getIconByTypeName(8, "automatic"),
                                blr.gfxResource.getIconByTypeName(6, "comfort"),
                                blr.gfxResource.getIconByTypeName(0, "standby"),
                                blr.gfxResource.getIconByTypeName(4, "economy"),
                                blr.gfxResource.getIconByTypeName(3, "protection")
                            ]
                        );

                        oItem = visitor.getComponentBy(
                            "knx",
                            "combo",
                            {
                                "knx/write-address": vimarGroupToKnxGroup(feedbacks.address),
                                "knx/feedback-address": vimarGroupToKnxGroup(feedbacks.address + 1),
                                "knx/virtual-feedback": true
                            }
                        );
                        if (oItem == undefined) {
                            oItem = application.addComponent({
                                type: "combo",
                                gateway: gateways.knx,
                                name: strIndex + "(VF)," + strZone + "," + strRoom + "," + strLabel + " " + _D("Mode"),
                                "knx/write-address": vimarGroupToKnxGroup(feedbacks.address),
                                "knx/feedback-address": vimarGroupToKnxGroup(feedbacks.address + 1),
                                "knx/virtual-feedback": true,
                                "knx/functional-block": 8,
                                "values": [{
                                    "value": 0,
                                    "description": _D("Automatic")
                                }, {
                                    "value": 1,
                                    "description": _D("Comfort")
                                }, {
                                    "value": 2,
                                    "description": _D("Standby")
                                }, {
                                    "value": 3,
                                    "description": _D("Economy")
                                }, {
                                    "value": 4,
                                    "description": _D("Building Protection")
                                }],
                                "icons": [
                                    blr.gfxResource.getIconByTypeName(0, "automatic"),
                                    blr.gfxResource.getIconByTypeName(1, "comfort"),
                                    blr.gfxResource.getIconByTypeName(2, "standby"),
                                    blr.gfxResource.getIconByTypeName(3, "economy"),
                                    blr.gfxResource.getIconByTypeName(4, "protection")
                                ]
                            }, groups.clima);
                        }

                        csv.data += build_csv_row(oItem.getData().getParamVal("name"), "20.102", vimarGroupToKnxGroup(feedbacks.address), vimarGroupToKnxGroup(feedbacks.address + 1), "");
                        feedbacks.address += 2;
                    }

                    // Valve state
                    oItem = visitor.getComponentBy(
                        "knx",
                        "info-text",
                        {
                            "knx/write-address": vimarGroupToKnxGroup(strGroupAddress),
                            "knx/feedback-address": vimarGroupToKnxGroup(strGroupAddress),
                            "knx/virtual-feedback": false,
                            "knx/read-init":false
                        }
                    );
                    if (oItem == undefined) {
                        oItem = application.addComponent({
                            type: "info-text",
                            gateway: gateways.knx,
                            name: strIndex + "," + strZone + "," + strRoom + "," + strLabel + " " + _D("Valve State"),
                            "knx/write-address": vimarGroupToKnxGroup(strGroupAddress),
                            "knx/feedback-address": vimarGroupToKnxGroup(strGroupAddress),
                            "knx/virtual-feedback": false,
                            "knx/read-init":false,
                            "text-0": "Off",
                            "text-1": "On",
                            "color-0": "#333333",
                            "color-1": "#990000"
                        }, groups.clima);
                    }

                    if (feedbacks.create) {
                        csv.data += build_csv_row(
                            oItem.getData().getParamVal("name"),
                            "1.001",
                            vimarGroupToKnxGroup(strGroupAddress),
                            "",
                            ""
                        );
                    }

                } break;
                // By-Me Fan-coil thermostat
                case 0x90: {
                    // Find lower group addresses as component group address 
                    var aGroupAddresses = new Array();
                    for (var i = 0; i < xmlDatapoints.length; ++i) {
                        if (xmlDatapoints[i].childNodes.length > 0) {
                            var xmlGroupObjects = xmlDatapoints[i].getElementsByTagName("groupobject");
                            if (xmlGroupObjects != undefined) {
                                for (var g = 0; g < xmlGroupObjects.length; ++g) {
                                    aGroupAddresses.push(xmlGroupObjects[g].getAttribute("address"));
                                }
                            }
                        }
                    }
                    var strGroupAddress = aGroupAddresses.sort()[0];

                    // Ambient Temperature
                    oItem = visitor.getComponentBy(
                        "vimar",
                        "info",
                        {
                            "vimar/group-address": strGroupAddress,
                            "vimar/physical-address": strHexAddress,
                            "vimar/device-pid": hexType.toString(16).toUpperCase(),
                            "vimar/device-fid": 12
                        }
                    );
                    if (oItem == undefined) {
                        oItem = application.addComponent({
                            type: "info",
                            gateway: gateways.vimar,
                            name: strIndex + "," + strZone + "," + strRoom + "," + strLabel + " " + _D("Ambient Temperature"),
                            "vimar/group-address": strGroupAddress,
                            "vimar/physical-address": strHexAddress,
                            "vimar/device-pid": hexType.toString(16).toUpperCase(),
                            "vimar/device-fid": 12
                        }, groups.clima);
                    }

                    if (feedbacks.create) {
                        oItem = visitor.getComponentBy(
                            "knx",
                            "info",
                            {
                                "knx/write-address": vimarGroupToKnxGroup(feedbacks.address),
                                "knx/feedback-address": vimarGroupToKnxGroup(feedbacks.address),
                                "knx/virtual-feedback": true,
                                "knx/raw-value": true,
                                "knx/DPT": "9.1"
                            }
                        );
                        if (oItem == undefined) {
                            oItem = application.addComponent({
                                type: "info",
                                gateway: gateways.knx,
                                name: strIndex + "(VF)," + strZone + "," + strRoom + "," + strLabel + " " + _D("Ambient Temperature"),
                                "knx/write-address": vimarGroupToKnxGroup(feedbacks.address),
                                "knx/feedback-address": vimarGroupToKnxGroup(feedbacks.address),
                                "knx/virtual-feedback": true,
                                "knx/raw-value": true,
                                "knx/DPT": "9.1"
                            }, groups.clima);
                        }

                        csv.data += build_csv_row(oItem.getData().getParamVal("name"), "9.001", "", vimarGroupToKnxGroup(feedbacks.address), "");
                        feedbacks.address += 1;
                    }

                    // Set point
                    oItem = visitor.getComponentBy(
                        "vimar",
                        "thermostat",
                        {
                            "vimar/group-address": strGroupAddress,
                            "vimar/physical-address": strHexAddress,
                            "vimar/device-pid": hexType.toString(16).toUpperCase(),
                            "vimar/device-fid": 6
                        }
                    );
                    if (oItem == undefined) {
                        oItem = application.addComponent({
                            type: "thermostat",
                            gateway: gateways.vimar,
                            name: strIndex + "," + strZone + "," + strRoom + "," + strLabel,
                            "vimar/group-address": strGroupAddress,
                            "vimar/physical-address": strHexAddress,
                            "vimar/device-pid": hexType.toString(16).toUpperCase(),
                            "vimar/device-fid": 6,
                            "icons": [
                                blr.gfxResource.getIconByTypeName(0, "temperature"),
                                blr.gfxResource.getIconByTypeName(1, "less"),
                                blr.gfxResource.getIconByTypeName(2, "circle"),
                                blr.gfxResource.getIconByTypeName(3, "plus")
                            ]
                        }, groups.clima);
                    }

                    if (feedbacks.create) {
                        oItem = visitor.getComponentBy(
                            "knx",
                            "thermostat",
                            {
                                "knx/write-address": vimarGroupToKnxGroup(feedbacks.address),
                                "knx/feedback-address": vimarGroupToKnxGroup(feedbacks.address + 1),
                                "knx/virtual-feedback": true
                            }
                        );
                        if (oItem == undefined) {
                            oItem = application.addComponent({
                                type: "thermostat",
                                gateway: gateways.knx,
                                name: strIndex + "(VF)," + strZone + "," + strRoom + "," + strLabel,
                                "knx/write-address": vimarGroupToKnxGroup(feedbacks.address),
                                "knx/feedback-address": vimarGroupToKnxGroup(feedbacks.address + 1),
                                "knx/virtual-feedback": true
                            }, groups.clima);
                        }

                        csv.data += build_csv_row(oItem.getData().getParamVal("name"), "9.001", vimarGroupToKnxGroup(feedbacks.address), vimarGroupToKnxGroup(feedbacks.address + 1), "");
                        feedbacks.address += 2;
                    }

                    // Crono
                    oItem = visitor.getComponentBy(
                        "vimar",
                        "timed-thermostat",
                        {
                            "vimar/group-address": strGroupAddress,
                            "vimar/physical-address": strHexAddress,
                            "vimar/device-pid": hexType.toString(16).toUpperCase(),
                            "vimar/device-fid": 6
                        }
                    );
                    if (oItem == undefined) {
                        oItem = application.addComponent({
                            type: "timed-thermostat",
                            gateway: gateways.vimar,
                            name: strIndex + "," + strZone + "," + strRoom + "," + strLabel + " " + _D("Crono"),
                            "vimar/group-address": strGroupAddress,
                            "vimar/physical-address": strHexAddress,
                            "vimar/device-pid": hexType.toString(16).toUpperCase(),
                            "vimar/device-fid": 6
                        }, groups.clima);
                    }

                    // Season
                    oItem = visitor.getComponentBy(
                        "vimar",
                        "combo",
                        {
                            "vimar/group-address": strGroupAddress,
                            "vimar/physical-address": strHexAddress,
                            "vimar/device-pid": hexType.toString(16).toUpperCase(),
                            "vimar/device-fid": 7,
                        }
                    );
                    if (oItem == undefined) {
                        oItem = application.addComponent({
                            type: "combo",
                            gateway: gateways.vimar,
                            name: strIndex + "," + strZone + "," + strRoom + "," + strLabel + " " + _D("Season"),
                            "vimar/group-address": strGroupAddress,
                            "vimar/physical-address": strHexAddress,
                            "vimar/device-pid": hexType.toString(16).toUpperCase(),
                            "vimar/device-fid": 7,
                            "values": [{
                                "value": 0,
                                "description": _D("Winter")
                            }, {
                                "value": 1,
                                "description": _D("Summer")
                            }],
                            "icons": [
                                blr.gfxResource.getIconByTypeName(0, "winter"),
                                blr.gfxResource.getIconByTypeName(1, "summer")
                            ]
                        }, groups.clima);
                    }

                    if (feedbacks.create) {
                        // To be defined
                    }

                    // Mode
                    oItem = visitor.getComponentBy(
                        "vimar",
                        "combo",
                        {
                            "vimar/group-address": strGroupAddress,
                            "vimar/physical-address": strHexAddress,
                            "vimar/device-pid": hexType.toString(16).toUpperCase(),
                            "vimar/device-fid": 1,
                        }
                    );
                    if (oItem == undefined) {
                        oItem = application.addComponent({
                            type: "combo",
                            gateway: gateways.vimar,
                            name: strIndex + "," + strZone + "," + strRoom + "," + strLabel + " " + _D("Mode"),
                            "vimar/group-address": strGroupAddress,
                            "vimar/physical-address": strHexAddress,
                            "vimar/device-pid": hexType.toString(16).toUpperCase(),
                            "vimar/device-fid": 1,
                            "values": [{
                                "value": 0,
                                "description": _D("Off")
                            }, {
                                "value": 1,
                                "description": _D("Forced Off")
                            }, {
                                "value": 2,
                                "description": _D("Off Period")
                            }, {
                                "value": 3,
                                "description": _D("Antifreeze")
                            }, {
                                "value": 4,
                                "description": _D("Night")
                            }, {
                                "value": 5,
                                "description": _D("Night Period")
                            }, {
                                "value": 6,
                                "description": _D("Manual")
                            }, {
                                "value": 7,
                                "description": _D("Manual Period")
                            }, {
                                "value": 8,
                                "description": _D("Automatic")
                            }],
                            "icons": [
                                blr.gfxResource.getIconByTypeName(0, "off"),
                                blr.gfxResource.getIconByTypeName(1, "forced"),
                                blr.gfxResource.getIconByTypeName(2, "off"),
                                blr.gfxResource.getIconByTypeName(3, "antifreeze"),
                                blr.gfxResource.getIconByTypeName(4, "night"),
                                blr.gfxResource.getIconByTypeName(5, "night"),
                                blr.gfxResource.getIconByTypeName(6, "manual"),
                                blr.gfxResource.getIconByTypeName(7, "manual"),
                                blr.gfxResource.getIconByTypeName(8, "automatic")
                            ]
                        }, groups.clima);
                    }

                    if (feedbacks.create) {
                        // Apply standard compatibility
                        oItem.getData().setParamVal(
                            "values",
                            [
                                {
                                    "value": 8,
                                    "description": _D("Automatic")
                                },
                                {
                                    "value": 6,
                                    "description": _D("Manual")
                                },
                                {
                                    "value": 0,
                                    "description": _D("Off")
                                },
                                {
                                    "value": 4,
                                    "description": _D("Night")
                                },
                                {
                                    "value": 3,
                                    "description": _D("Antifreeze")
                                }
                            ]
                        );

                        oItem.getData().setParamVal(
                            "icons",
                            [
                                blr.gfxResource.getIconByTypeName(8, "automatic"),
                                blr.gfxResource.getIconByTypeName(6, "comfort"),
                                blr.gfxResource.getIconByTypeName(0, "standby"),
                                blr.gfxResource.getIconByTypeName(4, "economy"),
                                blr.gfxResource.getIconByTypeName(3, "protection")
                            ]
                        );

                        oItem = visitor.getComponentBy(
                            "knx",
                            "combo",
                            {
                                "knx/write-address": vimarGroupToKnxGroup(feedbacks.address),
                                "knx/feedback-address": vimarGroupToKnxGroup(feedbacks.address + 1),
                                "knx/virtual-feedback": true
                            }
                        );
                        if (oItem == undefined) {
                            oItem = application.addComponent({
                                type: "combo",
                                gateway: gateways.knx,
                                name: strIndex + "(VF)," + strZone + "," + strRoom + "," + strLabel + " " + _D("Mode"),
                                "knx/write-address": vimarGroupToKnxGroup(feedbacks.address),
                                "knx/feedback-address": vimarGroupToKnxGroup(feedbacks.address + 1),
                                "knx/virtual-feedback": true,
                                "knx/functional-block": 8,
                                "values": [{
                                    "value": 0,
                                    "description": _D("Automatic")
                                }, {
                                    "value": 1,
                                    "description": _D("Comfort")
                                }, {
                                    "value": 2,
                                    "description": _D("Standby")
                                }, {
                                    "value": 3,
                                    "description": _D("Economy")
                                }, {
                                    "value": 4,
                                    "description": _D("Building Protection")
                                }],
                                "icons": [
                                    blr.gfxResource.getIconByTypeName(0, "automatic"),
                                    blr.gfxResource.getIconByTypeName(1, "comfort"),
                                    blr.gfxResource.getIconByTypeName(2, "standby"),
                                    blr.gfxResource.getIconByTypeName(3, "economy"),
                                    blr.gfxResource.getIconByTypeName(4, "protection")
                                ]
                            }, groups.clima);
                        }

                        csv.data += build_csv_row(oItem.getData().getParamVal("name"), "20.102", vimarGroupToKnxGroup(feedbacks.address), vimarGroupToKnxGroup(feedbacks.address + 1), "");
                        feedbacks.address += 2;
                    }

                    // Speed
                    oItem = visitor.getComponentBy(
                        "vimar",
                        "combo",
                        {
                            "vimar/group-address": strGroupAddress,
                            "vimar/physical-address": strHexAddress,
                            "vimar/device-pid": hexType.toString(16).toUpperCase(),
                            "vimar/device-fid": 2,
                        }
                    );
                    if (oItem == undefined) {
                        oItem = application.addComponent({
                            type: "combo",
                            gateway: gateways.vimar,
                            name: strIndex + "," + strZone + "," + strRoom + "," + strLabel + " " + _D("Speed"),
                            "vimar/group-address": strGroupAddress,
                            "vimar/physical-address": strHexAddress,
                            "vimar/device-pid": hexType.toString(16).toUpperCase(),
                            "vimar/device-fid": 2,
                            "values": [{
                                "value": 1,
                                "description": _D("Speed") + "1"
                            }, {
                                "value": 2,
                                "description": _D("Speed") + "2"
                            }, {
                                "value": 3,
                                "description": _D("Speed") + "3"
                            }, {
                                "value": 4,
                                "description": _D("Automatic")
                            }],
                            "icons": [
                                blr.gfxResource.getIconByTypeName(1, "speed_1"),
                                blr.gfxResource.getIconByTypeName(2, "speed_2"),
                                blr.gfxResource.getIconByTypeName(3, "speed_3"),
                                blr.gfxResource.getIconByTypeName(4, "automatic")
                            ]
                        }, groups.clima);
                    }

                    // speed result
                    oItem = visitor.getComponentBy(
                        "knx",
                        "combo",
                        {
                            "knx/write-address": vimarGroupToKnxGroup(strGroupAddress),
                            "knx/feedback-address": vimarGroupToKnxGroup(strGroupAddress),
                            "knx/functional-block": 8
                        }
                    );
                    if (oItem == undefined) {
                        oItem = application.addComponent({
                            type: "combo",
                            gateway: gateways.knx,
                            name: strIndex + "," + strZone + "," + strRoom + "," + strLabel + " " + _D("Actual") + " " + _D("Speed"),
                            "cache":true,
                            "read-only":true,
                            "knx/read-init":false,
                            "knx/write-address": vimarGroupToKnxGroup(strGroupAddress),
                            "knx/feedback-address": vimarGroupToKnxGroup(strGroupAddress),
                            "knx/functional-block": 8,
                            "values": [{
                                "value": 0,
                                "description": _D("Off")
                            }, {
                                "value": 65,
                                "description": _D("Cold") + " " + _D("Speed") + "1"
                            }, {
                                "value": 66,
                                "description": _D("Cold") + " " + _D("Speed") + "2"
                            }, {
                                "value": 68,
                                "description": _D("Cold") + " " + _D("Speed") + "3"
                            }, {
                                "value": 129,
                                "description": _D("Hot") + " " + _D("Speed") + "1"
                            }, {
                                "value": 130,
                                "description": _D("Hot") + " " + _D("Speed") + "2"
                            }, {
                                "value": 132,
                                "description": _D("Hot") + " " + _D("Speed") + "3"
                            }],
                            "icons": [
                                blr.gfxResource.getIconByTypeName(0, "off"),
                                blr.gfxResource.getIconByTypeName(65, "speed_1"),
                                blr.gfxResource.getIconByTypeName(66, "speed_2"),
                                blr.gfxResource.getIconByTypeName(68, "speed_3"),
                                blr.gfxResource.getIconByTypeName(129, "speed_1"),
                                blr.gfxResource.getIconByTypeName(130, "speed_2"),
                                blr.gfxResource.getIconByTypeName(132, "speed_3")
                            ]
                        }, groups.clima);
                    }

                    if (feedbacks.create) {
                        csv.data += build_csv_row(
                            oItem.getData().getParamVal("name"),
                            "5.001",
                            vimarGroupToKnxGroup(strGroupAddress),
                            "",
                            ""
                        );
                    }

                } break;
                default: {
                    console.warn("Unknown clima device [" + hexType.toString(16).toUpperCase() + "]" + strIndex + "," + strZone + "," + strRoom + "," + strLabel);
                } break;
            }
        } catch (err) {
            blr.executionStack.cancel();
            console.error(NAME + " error\n" + err);
            alert(NAME + " error\n" + err);
        }
        return oItem;
}

function create_media_component
(
    visitor,
    xmlIndex,
    xmlDatapoints,
    xmlDevice,
    xmlFunctionalBlock,
    groups,
    gateways,
    scenaries,
    feedbacks,
    csv
) {
    var oItem = undefined;
    try {
        var strAddressesScenario = new Array();
        var strIndex = xmlIndex.getAttribute("index");
        var strLabel = xmlIndex.getAttribute("label");
        var strRoom = xmlIndex.getAttribute("room");
        var strZone = xmlIndex.getAttribute("zone");
        var strHexAddress = xmlDevice.getAttribute("address");
        var hexType = parseInt(xmlDevice.getAttribute("type").toString(16), 16);
        var iFid = parseInt(xmlFunctionalBlock.getAttribute("index"), 10);

        // Update local scenaries
        var xmlFbGroupObjects = xmlFunctionalBlock.getElementsByTagName("groupobjects")[0];
        if (xmlFbGroupObjects != undefined) {
            var xmlFbDataType = xmlFbGroupObjects.getElementsByTagName("datapointtype");
            if (xmlFbDataType != undefined) {
                for (var i = 0; i < xmlFbDataType.length; ++i) {
                    var xmlDptType = xmlFbDataType[i];
                    var strName = xmlDptType.getAttribute("name");
                    if (strName == "DPTVimarScene") {
                        var xmlGroupObjects = xmlDptType.getElementsByTagName("groupobject");
                        if (xmlGroupObjects != undefined) {
                            for (var g = 0; g < xmlGroupObjects.length; ++g) {
                                var xmlGO = xmlGroupObjects[g];
                                var strAddressScenario = xmlGO.getAttribute("address");

                                if (scenaries[strAddressScenario] == undefined) {
                                    scenaries[strAddressScenario] = new Array();
                                }
                                scenaries[strAddressScenario].push(strHexAddress);
                                strAddressesScenario.push(strAddressScenario);
                            }
                        }
                    }
                }
            }
        }

        switch (hexType) {
            // Knx Media
            case 0x9F: {
                var playbackControlWAdr = 0;
                var playbackControlRAdr = 0;
                var playbackMenuWAdr = 0;
                for (var i = 0; i < xmlDatapoints.length; ++i) {
                    if (xmlDatapoints[i].childNodes.length > 0) {
                        var strName = xmlDatapoints[i].getAttribute("name");
                        var xmlGroupObjects = xmlDatapoints[i].getElementsByTagName("groupobject");
                        switch (strName) {
                            case "DPTCntrlPBStatus": {
                                playbackControlWAdr = parseInt(xmlGroupObjects[0].getAttribute("address"), 16);
                            } break;
                            case "DPTPBInfoStatus": {
                                playbackControlRAdr = parseInt(xmlGroupObjects[0].getAttribute("address"), 16);
                            } break;
                            case "DPTCntrlDBMenu": {
                                playbackMenuWAdr = parseInt(xmlGroupObjects[0].getAttribute("address"), 16);
                            } break;
                        }
                    }
                }

                var oAudioDockStation = application.addComponent({
                    type: "group",
                    name: strLabel + " " + _D("Docking Station")
                }, groups.media);

                /*
				1 Byte
				0 = up
				1 = down
				2 = left
				3 = right
				*/
                if (playbackMenuWAdr != 0) {
                    oItem = visitor.getComponentBy(
                        "knx",
                        "scenario",
                        {
                            "knx/write-address": vimarGroupToKnxGroup(playbackMenuWAdr),
                            "knx/functional-block": 8,
                            "knx/scene-value": 0,
                        }
                    );
                    if (oItem == undefined) {
                        oItem = application.addComponent({
                            type: "scenario",
                            gateway: gateways.knx,
                            name: strIndex + "," + strZone + "," + strRoom + "," + strLabel + " " + _D("Menu Up"),
                            "knx/write-address": vimarGroupToKnxGroup(playbackMenuWAdr),
                            "knx/functional-block": 8,
                            "knx/scene-value": 0,
                        }, oAudioDockStation);
                    }

                    csv.data += build_csv_row(
                        oItem.getData().getParamVal("name"),
                        "5.010",
                        vimarGroupToKnxGroup(playbackMenuWAdr),
                        "",
                        "0"
                    );

                    oItem = visitor.getComponentBy(
                        "knx",
                        "scenario",
                        {
                            "knx/write-address": vimarGroupToKnxGroup(playbackMenuWAdr),
                            "knx/functional-block": 8,
                            "knx/scene-value": 1,
                        }
                    );
                    if (oItem == undefined) {
                        oItem = application.addComponent({
                            type: "scenario",
                            gateway: gateways.knx,
                            name: strIndex + "," + strZone + "," + strRoom + "," + strLabel + " " + _D("Menu Down"),
                            "knx/write-address": vimarGroupToKnxGroup(playbackMenuWAdr),
                            "knx/functional-block": 8,
                            "knx/scene-value": 1,
                        }, oAudioDockStation);
                    }

                    csv.data += build_csv_row(
                        oItem.getData().getParamVal("name"),
                        "5.010",
                        vimarGroupToKnxGroup(playbackMenuWAdr),
                        "",
                        "1"
                    );

                    oItem = visitor.getComponentBy(
                        "knx",
                        "scenario",
                        {
                            "knx/write-address": vimarGroupToKnxGroup(playbackMenuWAdr),
                            "knx/functional-block": 8,
                            "knx/scene-value": 2,
                        }
                    );
                    if (oItem == undefined) {
                        oItem = application.addComponent({
                            type: "scenario",
                            gateway: gateways.knx,
                            name: strIndex + "," + strZone + "," + strRoom + "," + strLabel + " " + _D("Menu Left"),
                            "knx/write-address": vimarGroupToKnxGroup(playbackMenuWAdr),
                            "knx/functional-block": 8,
                            "knx/scene-value": 2,
                        }, oAudioDockStation);
                    }

                    csv.data += build_csv_row(
                        oItem.getData().getParamVal("name"),
                        "5.010",
                        vimarGroupToKnxGroup(playbackMenuWAdr),
                        "",
                        "2"
                    );

                    oItem = visitor.getComponentBy(
                        "knx",
                        "scenario",
                        {
                            "knx/write-address": vimarGroupToKnxGroup(playbackMenuWAdr),
                            "knx/functional-block": 8,
                            "knx/scene-value": 3,
                        }
                    );
                    if (oItem == undefined) {
                        oItem = application.addComponent({
                            type: "scenario",
                            gateway: gateways.knx,
                            name: strIndex + "," + strZone + "," + strRoom + "," + strLabel + " " + _D("Menu Right"),
                            "knx/write-address": vimarGroupToKnxGroup(playbackMenuWAdr),
                            "knx/functional-block": 8,
                            "knx/scene-value": 3,
                        }, oAudioDockStation);
                    }

                    csv.data += build_csv_row(
                        oItem.getData().getParamVal("name"),
                        "5.010",
                        vimarGroupToKnxGroup(playbackMenuWAdr),
                        "",
                        "3"
                    );
                }

                /*
				1 Byte
				0-1 bits:
				0 = nothing
				1 = loop current song
				2 = loop album

				2-3 bits:
				0 = nothing
				1 = shuffle song
				2 = shuffle album

				4 bit
				0 = pause
				1 = play

				5-6 bits:
				0 = ?
				1 = ?
				2 = prex
				3 = next

				7 = 0
				*/
                if (playbackControlWAdr != 0) {
                    oItem = visitor.getComponentBy(
                        "knx",
                        "scenario",
                        {
                            "knx/write-address": vimarGroupToKnxGroup(playbackControlWAdr),
                            "knx/functional-block": 8,
                            "knx/scene-value": 80,
                        }
                    );
                    if (oItem == undefined) {
                        oItem = application.addComponent({
                            type: "scenario",
                            gateway: gateways.knx,
                            name: strIndex + "," + strZone + "," + strRoom + "," + strLabel + " " + _D("Prev Track"),
                            "knx/write-address": vimarGroupToKnxGroup(playbackControlWAdr),
                            "knx/functional-block": 8,
                            "knx/scene-value": 80,
                        }, oAudioDockStation);
                    }

                    csv.data += build_csv_row(
                        oItem.getData().getParamVal("name"),
                        "5.010",
                        vimarGroupToKnxGroup(playbackControlWAdr),
                        "",
                        "80"
                    );

                    oItem = visitor.getComponentBy(
                        "knx",
                        "scenario",
                        {
                            "knx/write-address": vimarGroupToKnxGroup(playbackControlWAdr),
                            "knx/functional-block": 8,
                            "knx/scene-value": 112,
                        }
                    );
                    if (oItem == undefined) {
                        oItem = application.addComponent({
                            type: "scenario",
                            gateway: gateways.knx,
                            name: strIndex + "," + strZone + "," + strRoom + "," + strLabel + " " + _D("Next Track"),
                            "knx/write-address": vimarGroupToKnxGroup(playbackControlWAdr),
                            "knx/functional-block": 8,
                            "knx/scene-value": 112,
                        }, oAudioDockStation);
                    }

                    csv.data += build_csv_row(
                        oItem.getData().getParamVal("name"),
                        "5.010",
                        vimarGroupToKnxGroup(playbackControlWAdr),
                        "",
                        "112"
                    );
                }

                if (strAddressesScenario.length > 0) {// this object is inside a scenario
                    // We need to add the switchReadAdr and brightReadAdr addresses
                    for (var sa = 0; sa < strAddressesScenario.length; ++sa) {
                        var aData = scenaries[strAddressesScenario[sa]];
                        aData.pop();
                        // We just added
                        if (playbackControlRAdr != 0) { aData.push("G" + playbackControlRAdr); }
                    }
                }

            } break;
            case 0x9B: {
                var scanFrequencyWAdr = 0;
                var skipMemoryWAdr = 0;
                for (var i = 0; i < xmlDatapoints.length; ++i) {
                    if (xmlDatapoints[i].childNodes.length > 0) {
                        var strName = xmlDatapoints[i].getAttribute("name");
                        var xmlGroupObjects = xmlDatapoints[i].getElementsByTagName("groupobject");
                        switch (strName) {
                            case "DPTScanFrequency": {
                                scanFrequencyWAdr = parseInt(xmlGroupObjects[0].getAttribute("address"), 16);
                            } break;
                            case "DPTSkipMemoryFM": {
                                skipMemoryWAdr = parseInt(xmlGroupObjects[0].getAttribute("address"), 16);
                            } break;
                        }
                    }
                }

                var oAudioFM = application.addComponent({
                    type: "group",
                    name: strLabel + " " + _D("Radio FM")
                }, groups.media);

                // Scan Frequency
                if (scanFrequencyWAdr != 0) {
                    oItem = visitor.getComponentBy(
                        "knx",
                        "scenario",
                        {
                            "knx/write-address": vimarGroupToKnxGroup(scanFrequencyWAdr),
                            "knx/functional-block": 6,
                            "knx/scene-value": 0,
                        }
                    );
                    if (oItem == undefined) {
                        oItem = application.addComponent({
                            type: "scenario",
                            gateway: gateways.knx,
                            name: strIndex + "," + strZone + "," + strRoom + "," + strLabel + " " + _D("Scan Prev"),
                            "knx/write-address": vimarGroupToKnxGroup(scanFrequencyWAdr),
                            "knx/functional-block": 6,
                            "knx/scene-value": 0,
                        }, oAudioFM);
                    }

                    csv.data += build_csv_row(
                        oItem.getData().getParamVal("name"),
                        "1.022",
                        vimarGroupToKnxGroup(scanFrequencyWAdr),
                        "",
                        "0"
                    );

                    oItem = visitor.getComponentBy(
                        "knx",
                        "scenario",
                        {
                            "knx/write-address": vimarGroupToKnxGroup(scanFrequencyWAdr),
                            "knx/functional-block": 6,
                            "knx/scene-value": 1,
                        }
                    );
                    if (oItem == undefined) {
                        oItem = application.addComponent({
                            type: "scenario",
                            gateway: gateways.knx,
                            name: strIndex + "," + strZone + "," + strRoom + "," + strLabel + " " + _D("Scan next"),
                            "knx/write-address": vimarGroupToKnxGroup(scanFrequencyWAdr),
                            "knx/functional-block": 6,
                            "knx/scene-value": 1,
                        }, oAudioFM);
                    }

                    csv.data += build_csv_row(
                        oItem.getData().getParamVal("name"),
                        "1.022",
                        vimarGroupToKnxGroup(scanFrequencyWAdr),
                        "",
                        "1"
                    );
                }
                if (skipMemoryWAdr != 0) {
                    oItem = visitor.getComponentBy(
                        "knx",
                        "scenario",
                        {
                            "knx/write-address": vimarGroupToKnxGroup(skipMemoryWAdr),
                            "knx/functional-block": 6,
                            "knx/scene-value": 0,
                        }
                    );
                    if (oItem == undefined) {
                        oItem = application.addComponent({
                            type: "scenario",
                            gateway: gateways.knx,
                            name: strIndex + "," + strZone + "," + strRoom + "," + strLabel + " " + _D("Memory Prev"),
                            "knx/write-address": vimarGroupToKnxGroup(skipMemoryWAdr),
                            "knx/functional-block": 6,
                            "knx/scene-value": 0,
                        }, oAudioFM);
                    }

                    csv.data += build_csv_row(
                        oItem.getData().getParamVal("name"),
                        "1.022",
                        vimarGroupToKnxGroup(skipMemoryWAdr),
                        "",
                        "0"
                    );

                    oItem = visitor.getComponentBy(
                        "knx",
                        "scenario",
                        {
                            "knx/write-address": vimarGroupToKnxGroup(skipMemoryWAdr),
                            "knx/functional-block": 6,
                            "knx/scene-value": 1,
                        }
                    );
                    if (oItem == undefined) {
                        oItem = application.addComponent({
                            type: "scenario",
                            gateway: gateways.knx,
                            name: strIndex + "," + strZone + "," + strRoom + "," + strLabel + " " + _D("Memory next"),
                            "knx/write-address": vimarGroupToKnxGroup(skipMemoryWAdr),
                            "knx/functional-block": 6,
                            "knx/scene-value": 1,
                        }, oAudioFM);
                    }

                    csv.data += build_csv_row(
                        oItem.getData().getParamVal("name"),
                        "1.022",
                        vimarGroupToKnxGroup(skipMemoryWAdr),
                        "",
                        "1"
                    );
                }
            } break;
            case 0x9D:
            case 0x99: {
                var onOffAdr = 0;
                var volumeAdr = 0;
                var channelAdr = 0;
                for (var i = 0; i < xmlDatapoints.length; ++i) {
                    if (xmlDatapoints[i].childNodes.length > 0) {
                        var strName = xmlDatapoints[i].getAttribute("name");
                        var xmlGroupObjects = xmlDatapoints[i].getElementsByTagName("groupobject");
                        switch (strName) {
                            case "DPTOnOff": {
                                onOffAdr = parseInt(xmlGroupObjects[0].getAttribute("address"), 16);
                            } break;
                            case "DPT_Volume_Scaling": {
                                volumeAdr = parseInt(xmlGroupObjects[0].getAttribute("address"), 16);
                            } break;
                            case "DPTChannel": {
                                channelAdr = parseInt(xmlGroupObjects[0].getAttribute("address"), 16);
                            } break;
                        }
                    }
                }

                var oAudioZone = application.addComponent({
                    type: "group",
                    name: strIndex + "," + strZone + "," + strRoom + "," + strLabel
                }, groups.media);

                // on off
                if (onOffAdr != 0) {
                    oItem = visitor.getComponentBy(
                        "knx",
                        "switch",
                        {
                            "knx/write-address": vimarGroupToKnxGroup(onOffAdr),
                            "knx/feedback-address": vimarGroupToKnxGroup(onOffAdr),
                        }
                    );
                    if (oItem == undefined) {
                        oItem = application.addComponent({
                            type: "switch",
                            gateway: gateways.knx,
                            name: strIndex + "," + strZone + "," + strRoom + "," + strLabel + " " + _D("On - Off"),
                            "knx/write-address": vimarGroupToKnxGroup(onOffAdr),
                            "knx/feedback-address": vimarGroupToKnxGroup(onOffAdr),
                            "icons": [
                                blr.gfxResource.getIconByTypeName(0, "off"),
                                blr.gfxResource.getIconByTypeName(1, "on")
                            ]
                        }, oAudioZone);
                    }

                    csv.data += build_csv_row(
                        oItem.getData().getParamVal("name"),
                        "1.001",
                        vimarGroupToKnxGroup(onOffAdr),
                        vimarGroupToKnxGroup(onOffAdr),
                        ""
                    );
                }
                // Volume
                if (volumeAdr != 0) {
                    oItem = visitor.getComponentBy(
                        "knx",
                        "slider",
                        {
                            "knx/write-address": vimarGroupToKnxGroup(volumeAdr),
                            "knx/feedback-address": vimarGroupToKnxGroup(volumeAdr),
                        }
                    );
                    if (oItem == undefined) {
                        oItem = application.addComponent({
                            type: "slider",
                            gateway: gateways.knx,
                            name: strIndex + "," + strZone + "," + strRoom + "," + strLabel + " " + _D("Volume"),
                            "knx/write-address": vimarGroupToKnxGroup(volumeAdr),
                            "knx/feedback-address": vimarGroupToKnxGroup(volumeAdr),
                            "knx/range": 0,
                            "icons": [
                                blr.gfxResource.getIconByTypeName(0, "off"),
                                blr.gfxResource.getIconByTypeName(1, "on")
                            ]
                        }, oAudioZone);
                    }

                    csv.data += build_csv_row(
                        oItem.getData().getParamVal("name"),
                        "5.010",
                        vimarGroupToKnxGroup(volumeAdr),
                        vimarGroupToKnxGroup(volumeAdr),
                        "[0-127]"
                    );
                }
                // Channels
                /*
				 1 Byte
				 0 = DockStation
				 1 = Radio
				 3 = IR
				 */
                if (channelAdr != 0) {
                    oItem = visitor.getComponentBy(
                        "knx",
                        "combo",
                        {
                            "knx/write-address": vimarGroupToKnxGroup(channelAdr),
                            "knx/feedback-address": vimarGroupToKnxGroup(channelAdr),
                            "knx/range": 0,
                        }
                    );
                    if (oItem == undefined) {
                        oItem = application.addComponent({
                            type: "combo",
                            gateway: gateways.knx,
                            name: strIndex + "," + strZone + "," + strRoom + "," + strLabel + " " + _D("Sources"),
                            "knx/write-address": vimarGroupToKnxGroup(channelAdr),
                            "knx/feedback-address": vimarGroupToKnxGroup(channelAdr),
                            "knx/functional-block": 6,
                            "values": [{
                                "value": 0,
                                "description": _D("Docking Station")
                            }, {
                                "value": 1,
                                "description": _D("Radio")
                            }, {
                                "value": 3,
                                "description": _D("IR")
                            }],
                            "icons": [
                                blr.gfxResource.getIconByTypeName(0, "mp3"),
                                blr.gfxResource.getIconByTypeName(1, "tuner"),
                                blr.gfxResource.getIconByTypeName(3, "circle", 200, 0, 0)
                            ]
                        }, oAudioZone);
                    }

                    csv.data += build_csv_row(
                        oItem.getData().getParamVal("name"),
                        "2.006",
                        vimarGroupToKnxGroup(channelAdr),
                        vimarGroupToKnxGroup(channelAdr),
                        "0 (Docking Station),1 (Radio),3 (IR)"
                    );
                }

                if (strAddressesScenario.length > 0) {// this object is inside a scenario
                    // We need to add the switchReadAdr and brightReadAdr addresses
                    for (var sa = 0; sa < strAddressesScenario.length; ++sa) {
                        var aData = scenaries[strAddressesScenario[sa]];
                        aData.pop();
                        // We just added
                        if (onOffAdr != 0) { aData.push("G" + onOffAdr); }
                        if (volumeAdr != 0) { aData.push("G" + volumeAdr); }
                        if (channelAdr != 0) { aData.push("G" + channelAdr); }
                    }
                }

            } break;
            default: {
                console.warn("Unknown media device [" + hexType.toString(16).toUpperCase() + "]" + strIndex + "," + strZone + "," + strRoom + "," + strLabel);
            } break;
        }
    } catch (err) {
        blr.executionStack.cancel();
        console.error(NAME + " error\n" + err);
        alert(NAME + " error\n" + err);
    }
    return oItem;
}

function getGateways(gtwIp) {
    var gateways = {
        knx: undefined,
        vimar: undefined
    };

    // If already exist in the project take them, otherwise create
    var aKnx = application.getGatewaysByProtocol("knx");
    var aVimar = application.getGatewaysByProtocol("vimar");

    if (aKnx.length == 0) {
        gateways.knx = application.addGateway({
            name: "Knx",
            protocol: "knx"
        });
        gateways.knx.data.setParamVal("knx/address", gtwIp);
    }else{
        gateways.knx = aKnx[0];
    }

    if (aVimar.length == 0) {
        gateways.vimar = application.addGateway({
            name: "Vimar",
            protocol: "vimar"
        });
        gateways.vimar.data.setParamVal("knx/address", gtwIp);
    } else {
        gateways.vimar = aVimar[0];
    }


    return gateways;
}

function create_gui() {
    // Current Gateway if found...
    var aKnx = application.getGatewaysByProtocol("knx");
    var aVimar = application.getGatewaysByProtocol("vimar");

    var defaultIPAddress = undefined;
    if (aVimar.length > 0 && defaultIPAddress == undefined) {
        defaultIPAddress = aVimar[0].data.getParamVal("knx/address");
    }

    if (aKnx.length > 0 && defaultIPAddress == undefined) {
        defaultIPAddress = aVimar[0].data.getParamVal("knx/address");
    }

    if (defaultIPAddress == undefined) {
        defaultIPAddress = "192.168.1.250";
    }

    var wizard = $("<div id='vimar-byme-importer' title='" + NAME + "'></div>");
    wizard.append("<label>Database file WebServer (.xml):</label></br></br>");
    wizard.append("<button id='btn-open'>Select file</button>&nbsp;<span id='vimar-byme-file'></span></br></br>");

    wizard.append("<label>Gateway Address:</label></br></br>");
    wizard.append("<input type='text' id='gtw-address' value='" + defaultIPAddress + "'/></br></br>");

    wizard.append("<input id='create-feedbacks' type='checkbox' />&nbsp;<label>Create KNX virtual feedbacks</label></br></br>");

    wizard.append("&nbsp;<label>Starting KNX group address:</label></br></br>");
    wizard.append("&nbsp;<input type='text' id='knx-start-address' value='11/0/0'/></br></br>");

    wizard.append("<label>Reference CSV file folder:</label></br></br>");
    wizard.append("<button id='btn-save'>Select folder</button>&nbsp;<span id='knx-file'></span></br></br>");

    $("body").append(wizard);

    wizard.find("#btn-open").button().click(function (event) {
        var strFile = QT.getOpenFileName("Select Vimar By-Me xml file", "", "*.xml");
        wizard.find("#vimar-byme-file").text(strFile);
        event.preventDefault();
    });

    wizard.find("#btn-save").button().click(function (event) {
        var strFile = QT.getSaveFileName("Choose Where to save KNX csv file", "", "*.csv");
        if (strFile != "") {
            var suffix = ".csv";
            var endWith = strFile.substring(strFile.length - suffix.length);
            if (endWith != suffix) {
                strFile += suffix;
            }
            wizard.find("#knx-file").text(strFile);
        }
        event.preventDefault();
    });

    wizard.dialog({
        modal: true,
        width:"50%",
        close: function () {
            $(this).remove();
        },
        buttons: {
            "Import": function () {
                $(this).dialog("close");
                $(this).remove();
                
                /*
                console.log("Create gateway  ---- :" + $(this).children("#gtw-address").val());
                console.log("Read Vimar file ---- :" + $(this).children("#vimar-byme-file").text());
                console.log("Save Knx csv  ------ :" + $(this).children("#knx-file").text());
                console.log("Create feddbacks --- :" + $(this).children("#create-feedbacks")[0].checked);
                */

                blr.executionStack.show(NAME);
                blr.executionStack.push(
                    function (in_args, out_args) {
                        out_args = vimar_main(
                            in_args.gateway,
                            in_args.byme,
                            in_args.knx,
                            in_args.feedbacks,
                            in_args.startaddress
                        );
                    },
                    {
                        gateway: $(this).children("#gtw-address").val(),
                        byme: $(this).children("#vimar-byme-file").text(),
                        knx: $(this).children("#knx-file").text(),
                        feedbacks: $(this).children("#create-feedbacks")[0].checked,
                        startaddress: $(this).children("#knx-start-address").val()
                    }
                );
                blr.executionStack.execute(1, false, false);
            },
            Cancel: function () {
                $(this).dialog("close");
                $(this).remove();
            }
        }
    });

    return wizard;
}

function build_csv_row
(
    name,
    type,
    read,
    write,
    value
)
{
    return name + "\t" + type + "\t" + read + "\t" + write + "\t" + value + "\n";
}

/** Vimar Group Address Converter
 */
function vimarGroupToKnxGroup(nAdr) {
    var iVal = nAdr;
    if (typeof nAdr === 'string') {
        iVal = parseInt(nAdr, 16);
    }

    var A = (iVal >> 11) & 0x1F;

    var B = (iVal >> 8) & 0x07;

    var C = (iVal) & 0xFF;

    return A + "/" + B + "/" + C;
};

function groupAddressStringToValue(groupAddress) {
    if (typeof groupAddress === 'string') {
        var aValues = groupAddress.split("/");
        if (aValues.length < 3) {
            return -1;
        }

        var A = parseInt(aValues[0], 10);
        var B = parseInt(aValues[1], 10);
        var C = parseInt(aValues[2], 10);

        // Compose the result
        return ((A & 0x1F) << 11) | ((B & 0x07) << 8) | ((C & 0xFF));
    }

    return -1;
}

/** Vimar Physical Address Converter
 */
function vimarAddressToKnxAddress(nAdr) {
    var iVal = nAdr;
    if (typeof nAdr === 'string') {
        iVal = parseInt(nAdr, 16);
    }

    var A = ((iVal >> 12) & 0x0F);

    var B = ((iVal >> 8) & 0x0F);

    var C = ((iVal) & 0xFF);

    return A + "." + B + "." + C;
};

function physicalAddressStringToValue(physicalAddress) {
    if (typeof physicalAddress === 'string') {
        var aValues = physicalAddress.split(".");
        if (aValues.length < 3) {
            return -1;
        }

        var A = parseInt(aValues[0], 10);
        var B = parseInt(aValues[1], 10);
        var C = parseInt(aValues[2], 10);

        // Compose the result
        return ((A & 0x0F) << 12) | ((B & 0x0F) << 8) | ((C & 0xFF));
    }

    return -1;
}


function isBymeProduct(hexType) {
    switch (hexType) {
        case 0x60: { return true; } break;
        case 0x61: { return true; } break;
        case 0x62: { return true; } break;
        case 0x63: { return true; } break;
        case 0x64: { return true; } break;
        case 0x65: { return true; } break;
        case 0x66: { return true; } break;
        case 0x68: { return true; } break;
        case 0x69: { return true; } break;
        case 0x6A: { return true; } break;
        case 0x6B: { return true; } break;
        case 0x6C: { return true; } break;
        case 0x6D: { return true; } break;
        case 0x6E: { return true; } break;
        case 0x6F: { return true; } break;
        case 0x70: { return true; } break;
        case 0x73: { return true; } break;
        case 0x75: { return true; } break;
        case 0x7B: { return true; } break;
        case 0x7C: { return true; } break;
        case 0x82: { return true; } break;
        case 0x85: { return true; } break;
        case 0x86: { return true; } break;
        case 0x87: { return true; } break;
        case 0x90: { return true; } break;
        case 0x91: { return true; } break;
        case 0xA1: { return true; } break;
        case 0xA2: { return true; } break;
        case 0xA3: { return true; } break;
            // Know Knx Devices

    }
    return false;
}

